# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

"""
This module exports :class:`auromat.mapping.mapping.BaseMapping` objects
into the netCDF file format following the CF 1.6 and NODC conventions.
Each mapping is exported as a single self-contained file.

See http://cfconventions.org/1.6.html,
http://www.nodc.noaa.gov/data/formats/netcdf/grid.cdl and
http://wiki.esipfed.org/index.php/NetCDF-CF_File_Examples_for_Satellite_Swath_Data
for details.
"""

import sys
from datetime import datetime
import numpy as np
from netCDF4 import Dataset

from auromat.coordinates.transform import northGeomagneticPoleLocation
from auromat.mapping.mapping import isPlateCarree

def write(outputPath, mapping, metadata={}, includeBounds=True,
          includeMagCoords=True, includeGeoCoords=True,
          use1dIfPossible=True, compress=True):
    """
    
    :param str outputPath:
    :param auromat.mapping.mapping.BaseMapping mapping:
    :param dict metadata: additional metadata, overwrites mapping.metadata entries if existing,
                          a dictionary of root attributes, e.g.::
                          
                            {'Project': '..',
                             'Source_name': '..',
                             'Discipline': 'Space Physics>Magnetospheric Science',
                             'Descriptor': '..'
                            }
                            
                          See http://wiki.esipfed.org/index.php/Attribute_Convention_for_Data_Discovery.
    :param bool includeBounds: stores the coordinates of each pixel corner (in addition to the center)
    :param bool includeMagCoords: include geomagnetic latitude-magnetic local time coordinates
    :param bool compress: use compression for variables
    """
    if not includeGeoCoords:
        raise ValueError('Geodetic coordinates cannot be disabled for netCDF as they are essential to the format')
    
    with Dataset(outputPath, 'w', format='NETCDF4') as root:
        # ROOT ATTRIBUTES
        root.Conventions = 'CF-1.6'
        
        metadata = dict(list(mapping.metadata.items()) + list(metadata.items()))
        for k,v in metadata.items():
            if isinstance(v, bool):
                v = np.uint8(v)
            try:
                setattr(root, k, v)
            except TypeError:
                print('Cannot store global attribute "{}" with value {}'.format(k,repr(v)), file=sys.stderr) 
                raise
            
        isLatLonPlateCarree = use1dIfPossible and isPlateCarree(mapping.lats, mapping.lons)
        isMLatMltPlateCarree = use1dIfPossible and isPlateCarree(*mapping.mLatMlt) if includeMagCoords else None
        
        # NODC conventions:
        root.geospatial_lat_min = mapping.boundingBox.latSouth
        root.geospatial_lat_max = mapping.boundingBox.latNorth
        root.geospatial_lon_min = mapping.boundingBox.lonWest
        root.geospatial_lon_max = mapping.boundingBox.lonEast
        root.geospatial_lat_units = 'degrees_north'
        root.geospatial_lon_units = 'degrees_east'
                
        # DIMENSIONS
        # NOTE There is a bug preventing that a dimension is named equally as a variable.
        #      Therefore we append an 's' to the dimension names where it matters.
        w = mapping.img.shape[1]
        h = mapping.img.shape[0]
        if isLatLonPlateCarree:
            root.createDimension('lats', h)
            root.createDimension('lons', w)
        if isMLatMltPlateCarree:
            root.createDimension('mlats', h)
            root.createDimension('mlts', w)
        if not isLatLonPlateCarree or isMLatMltPlateCarree is False: # leave as "is False"! see above
            root.createDimension('y', h)
            root.createDimension('x', w)
        if includeBounds:
            if isLatLonPlateCarree or isMLatMltPlateCarree:
                root.createDimension('vertex2', 2) # for 1D coordinate arrays
            if not isLatLonPlateCarree or isMLatMltPlateCarree is False:
                root.createDimension('vertex4', 4) # for 2D coordinate arrays        
        root.createDimension('channel', mapping.img.shape[2])
        root.createDimension('xyz', 3)
        
        # COORDINATE VARIABLES
        
        # It seems that netcdf-CF defines time as in POSIX timestamps (without leap seconds).
        # see https://github.com/Unidata/netcdf4-python/issues/280
        # This means that a dataset covering a leap second with 1 second resolution will have
        # one data point appearing twice when the leap second happens.
        # Supporting leap seconds would require more effort though, as Python's datetime has the
        # same problem, not to mention what camera EXIF times would be like.
        time = root.createVariable('time', np.float64)
        time.units = 'seconds since 1970-01-01 00:00:00'
        time.calendar = 'gregorian'
        time.standard_name = 'time'
        time.axis = 'T'
        time.long_name = ''
        time.comment = ''
        time[:] = _unix(mapping.photoTime)
        # Note: Adding time bounds using the EXIF exposure time doesn't make sense here.
        #       The reason is that this would require very accurate EXIF times from
        #       the beginning. For the ISS mappings, this is not the case.
        
        # Note that chunk sizes are chosen such that they are compatible to GDAL
        # see http://trac.osgeo.org/gdal/ticket/4513#comment:18
        
        # Note that for all float variables we use NaN to indicate missing values (instead of _FillValue).
        # see http://www.unidata.ucar.edu/software/netcdf/docs/BestPractices.html
                     
        # lat lon coordinates have a time dimension here as each non-resampled mapping has its lat lon arrays.
        if isLatLonPlateCarree:
            # Note that we use the unmasked arrays here to stay compatible with the CF conventions
            # which say that coordinate arrays should not have missing values. This also makes it more compatible
            # with standard software.
            latsCenter, lonsCenter = mapping.latsCenter.data[:,0], mapping.lonsCenter.data[0,:]
            
            lat = root.createVariable('lat', np.float64, ('lats'), zlib=compress)
            lat.actual_range = np.float64([latsCenter[-1], latsCenter[0]])
            lat[:] = latsCenter
            
            lon = root.createVariable('lon', np.float64, ('lons'), zlib=compress)
            lon.actual_range = np.float64([lonsCenter[0], lonsCenter[-1]])
            lon[:] = lonsCenter
            
            # Note that plate carree is not yet defined as a recognized projection within CF. 
            # Therefore we cannot use a more specific crs.grid_mapping_name value.
            
        else:
            # auxiliary 2D coordinate variables with non-standard projection
            # see http://cfconventions.org/1.6.html#idp5561056  
            
            # IMPORTANT:
            # netCDF-CF 1.6 does not allow missing values for coordinate arrays.
            # There is a proposal to add support for it in the future though:
            # https://cf-pcmdi.llnl.gov/trac/ticket/85
            # This means that we violate the CF conventions slightly, but as our irregular (curvilinear)
            # grid is not defined for all image pixels there is no alternative here (at least for
            # the original non-resampled mappings).
            
            lat = root.createVariable('lat', np.float64, ('y', 'x'), zlib=compress, chunksizes=(1, w))
            lat.actual_range = np.float64([np.min(mapping.latsCenter), np.max(mapping.latsCenter)])
            lat[:] = mapping.latsCenter.filled(np.nan)
            
            lon = root.createVariable('lon', np.float64, ('y', 'x'), zlib=compress, chunksizes=(1, w))
            lon.actual_range = np.float64([np.min(mapping.lonsCenter), np.max(mapping.lonsCenter)])
            lon[:] = mapping.lonsCenter.filled(np.nan)
            
        lat.units = 'degrees_north'
        lat.valid_min = np.float64(-90)
        lat.valid_max = np.float64(90)
        lat.standard_name = 'latitude'
        lat.axis = 'Y'
        lat.long_name = 'Latitude'
        lat.comment = 'Geodetic latitude'
        
        lon.units = 'degrees_east'
        lon.valid_min = np.float64(-180)
        lon.valid_max = np.float64(180)
        lon.standard_name = 'longitude'
        lon.axis = 'X'
        lon.long_name = 'Longitude'
        lon.comment = 'Geodetic longitude'
                
        altitude = root.createVariable('altitude', np.int32)
        altitude.units = 'meters'
        altitude.standard_name = 'height_above_reference_ellipsoid'
        altitude.axis = 'Z'
        altitude.long_name = ''
        #altitude.comment = ''
        altitude[:] = mapping.altitude * 1000

        if includeBounds:
            # http://cfconventions.org/1.6.html#cell-boundaries
            lat.bounds = 'lat_bounds'
            lon.bounds = 'lon_bounds'
            
            if isLatLonPlateCarree:
                lat_bounds = root.createVariable('lat_bounds', np.float64, ('lats', 'vertex2'), 
                                                 zlib=compress, chunksizes=(h, 2))
                lat_bounds[:] = _bounds1d(mapping.lats.data[:,0])
                                
                lon_bounds = root.createVariable('lon_bounds', np.float64, ('lons', 'vertex2'),
                                                 zlib=compress, chunksizes=(w, 2))
                lon_bounds[:] = _bounds1d(mapping.lons.data[0,:])
            else:         
                lat_bounds = root.createVariable('lat_bounds', np.float64, ('y', 'x', 'vertex4'), 
                                                 zlib=compress, chunksizes=(1, w, 4))
                lat_bounds[:] = _bounds2d(mapping.lats.filled(np.nan))
                
                lon_bounds = root.createVariable('lon_bounds', np.float64, ('y', 'x', 'vertex4'),
                                                 zlib=compress, chunksizes=(1, w, 4))
                lon_bounds[:] = _bounds2d(mapping.lons.filled(np.nan))
        
        if includeMagCoords:
            # The CF 1.6 standard has no convention to define coordinates in different systems and
            # somehow relate them to the existing primary coordinates.
            # Therefore the following is non-standard.
            
            mlats, mlts = mapping.mLatMltCenter
            
            if isMLatMltPlateCarree:
                mlatsCenter, mltsCenter = mlats.data[:,0], mlts.data[0,:]
                
                mlat = root.createVariable('mlat', np.float64, ('mlats'), zlib=compress)
                mlat.actual_range = np.float64([mlatsCenter[-1], mlatsCenter[0]])
                mlat[:] = mlatsCenter
                
                mlt = root.createVariable('mlt', np.float64, ('mlts'), zlib=compress)
                mlt.actual_range = np.float64([mltsCenter[0], mltsCenter[-1]])
                mlt[:] = mltsCenter
                
            else:  
                mlat = root.createVariable('mlat', np.float64, ('y', 'x'), zlib=compress, chunksizes=(1, w))
                mlat.actual_range =  np.float64([np.min(mlats), np.max(mlats)])
                mlat[:] = mlats.filled(np.nan)
                
                mlt = root.createVariable('mlt', np.float64, ('y', 'x'), zlib=compress, chunksizes=(1, w))
                mlt.actual_range = np.float64([np.min(mlts), np.max(mlts)])                
                mlt[:] = mlts.filled(np.nan)
              
            #mlat.standard_name = 'grid_latitude'
            mlat.long_name = 'Geomagnetic latitude'  
            mlat.units = 'degrees'
            mlat.valid_min = np.float64(-90)
            mlat.valid_max = np.float64(90)
            # There is no standard attribute to define the coordinate system on the coordinate
            # variables themselves. We use the non-standard 'crs' attribute for that purpose.
            mlat.crs = 'mcrs'
            
            mlt.long_name = 'Magnetic local time'
            mlt.units = 'hours'
            mlt.valid_min = np.float64(0)
            mlt.valid_max = np.float64(24)
            mlt.crs = 'mcrs'
                
            
            if includeBounds:
                mlat.bounds = 'mlat_bounds'
                mlt.bounds = 'mlt_bounds'
                
                mlats, mlts = mapping.mLatMlt
                
                if isMLatMltPlateCarree:
                    mlat_bounds = root.createVariable('mlat_bounds', np.float64, ('mlats', 'vertex2'), 
                                                      zlib=compress, chunksizes=(h, 2))
                    mlat_bounds[:] = _bounds1d(mlats.data[:,0])                
                    
                    mlt_bounds = root.createVariable('mlt_bounds', np.float64, ('mlts', 'vertex2'), 
                                                     zlib=compress, chunksizes=(w, 2))
                    mlt_bounds[:] = _bounds1d(mlts.data[0,:])
                else:                                
                    mlat_bounds = root.createVariable('mlat_bounds', np.float64, ('y', 'x', 'vertex4'), 
                                                      zlib=compress, chunksizes=(1, w, 4))
                    mlat_bounds[:] = _bounds2d(mlats.filled(np.nan))
                    
                    mlt_bounds = root.createVariable('mlt_bounds', np.float64, ('y', 'x', 'vertex4'), 
                                                     zlib=compress, chunksizes=(1, w, 4))
                    mlt_bounds[:] = _bounds2d(mlts.filled(np.nan))
            
            magPoleLat, magPoleLon = northGeomagneticPoleLocation(mapping.photoTime)
            
            # This is all non-standard.
            mcrs = root.createVariable('mcrs', np.int8) # holds no actual data
            mcrs.north_geomagnetic_pole_lat = magPoleLat
            mcrs.north_geomagnetic_pole_lon = magPoleLon
            mcrs.comment = 'Geocentric MLat/MLT system based on the given geomagnetic pole position'

        # DATA VARIABLES
        y = 'lats' if isLatLonPlateCarree else 'y'
        x = 'lons' if isLatLonPlateCarree else 'x'

        # netcdf doesn't support unsigned types except byte
        # we generically convert to int types and ignore the case of uint8 with no masked pixels
        imgDtypeMap = {np.dtype(np.uint8): np.int16,
                       np.dtype(np.uint16): np.int32,
                       }
        if not mapping.img.dtype in imgDtypeMap.keys():
            raise NotImplementedError('Image data type not supported: ' + str(mapping.img.dtype))
        
        imgDtype = imgDtypeMap[mapping.img.dtype]
        imgFillval = imgDtype(np.iinfo(imgDtype).min)
        img_ = mapping.img.astype(imgDtype).filled(imgFillval)
        
        if img_.shape[2] == 1:
            bands = ['img']
        elif img_.shape[2] == 3:
            bands = ['img_red', 'img_green', 'img_blue']
        else:
            raise NotImplementedError
        
        for i, band in enumerate(bands):
            img = root.createVariable(band, imgDtype, (y, x), fill_value=imgFillval, 
                                      zlib=compress, chunksizes=(1, w))
            img.units = 'unitless'
    
            # cell_methods is not set for img as it is too meaningless and ambiguous.
            # It is not clear whether the value of a pixel is really the 'mean' or
            # something else over time and area.
            # 'area' (equally 'x: y:') depends on the camera sensor type and the method for
            # generating pixels out of raw sensor data. E.g. for Bayer-filter cameras
            # demosaicing algorithms interpolate the pixel colours.
            img.valid_min = imgDtype(np.iinfo(mapping.img.dtype).min)
            img.valid_max = imgDtype(np.iinfo(mapping.img.dtype).max)
            img.actual_range =  imgDtype([np.min(mapping.img[:,:,i]), np.max(mapping.img[:,:,i])])

            #img.long_name = ''
            img.coordinates = 'altitude time' if isLatLonPlateCarree else 'lat lon altitude time'
            img.grid_mapping = 'crs'
            img[:] = img_[:,:,i]
        
        # netCDF-CF doesn't know elevation angle, but zenith angle
        # so we use that instead to follow the conventions
        zena = 90 - mapping.elevation
        zenith_angle = root.createVariable('zenith_angle', np.float32, (y, x), 
                                           zlib=compress, chunksizes=(1, w))
        zenith_angle.units = 'degrees'
        zenith_angle.cell_methods = 'time: lat: lon: point' if isLatLonPlateCarree else 'time: y: x: point'
        zenith_angle.valid_min = np.float32(0)
        zenith_angle.valid_max = np.float32(90)
        zenith_angle.actual_range = [np.min(zena), np.max(zena)]
        # zenith_angle: 0=zenith, 90=horizon
        zenith_angle.standard_name = 'zenith_angle'
        zenith_angle.coordinates = 'altitude time' if isLatLonPlateCarree else 'lat lon altitude time'
        zenith_angle.grid_mapping = 'crs'
        zenith_angle.long_name = 'Absolute sensor zenith angle'
        zenith_angle[:] = zena.filled(np.nan)
        
        cameraPos = root.createVariable('camera_pos', np.float64, ('xyz'))
        cameraPos.units = 'kilometers'
        cameraPos.cell_methods = 'time: point'
        cameraPos.coordinates = 'time'
        cameraPos.long_name = 'Camera position in cartesian GCRS coordinates'
        cameraPos.comment = 'Axis order: xyz'
        cameraPos[:] = mapping.cameraPosGCRS
        
        crs = root.createVariable('crs', np.int8) # holds no actual data
        crs.grid_mapping_name = 'latitude_longitude' # = unknown projection lat/lon coordinate system
        crs.semi_major_axis = 6378137.0
        crs.inverse_flattening = 298.257223563
        crs.comment = 'Geographic Coordinate System, WGS 84'

def _bounds1d(arr):
    assert arr.ndim == 1
    arr = arr[:,None]
    bounds = np.concatenate((arr[:-1],arr[1:]), axis=1)
    assert bounds.shape == (arr.shape[0]-1, 2)
    return bounds

def _bounds2d(arr):
    assert arr.ndim == 2
    arr = arr[:,:,None]
    bounds = np.concatenate((
                arr[0:-1, 0:-1],
                arr[0:-1, 1:  ],
                arr[1:  , 1:  ],
                arr[1:  , 0:-1],
                ), axis=2)
    assert bounds.shape == (arr.shape[0]-1, arr.shape[1]-1, 4)
    return bounds

def _unix(dt):
    return (dt - datetime(1970, 1, 1)).total_seconds()

if __name__ == '__main__':
    import os
    from auromat.mapping.spacecraft import getMapping
    
    m = getMapping('/home/mriecher/data/arrrgh/img/ISS030-E/102100/ISS030-E-102170.jpg',
                   '/home/mriecher/data/arrrgh/wcs/ISS030-E/102100/ISS030-E-102170_1727.wcs')

    path = '/user_data/mriecher/test.nc'
    if os.path.exists(path):
        os.remove(path)

    write(path, m, includeBounds=True, includeMagCoords=True, zlib=True)
    


"""
The mapping package is the heart of this library. 
Its functionality can be broadly split up in three areas:

- (Down-)Loading images and calibration data from different providers
- Calculation of geodetic coordinates and elevation angles at arbitrary altitude 
  for an image from calibration data
- Accessing derived information about a georeferenced image 
  (e.g. bounding box and outline, centroid, pixel resolution)
- Transformation of coordinates (masking, resampling, coordinate system change)

The calibration data kinds currently supported are:

- `FITS WCS <http://fits.gsfc.nasa.gov/fits_wcs.html>`_ (header-only) 
  with camera position and image timestamp
- `MIRACLE <http://www.space.fmi.fi/MIRACLE/ASC/index.html>`_ calibration text files
- raw coordinate arrays (from memory, or CDF/netCDF files like 
  `THEMIS <http://themis.ssl.berkeley.edu>`_)

The data providers currently supported are:

- FITS WCS file provider (local only)
- `ESA ISS aurora archive <http://cosmos.esa.int/arrrgh>`_ 
  provider (web access, uses the FITS WCS file provider)
- THEMIS provider (web access)
- MIRACLE provider (local only)
- CDF file provider (local only, format as in :mod:`auromat.export.cdf`)
- netCDF file provider (local only, format as in :mod:`auromat.export.netcdf`)

The starting point is the :mod:`auromat.mapping.mapping` module on which all other
modules depend. 
"""


# Copyright European Space Agency, 2013

"""
This module provides functions to extract camera and lens information from
EXIF data and correct lens distortion using the 
`lensfunpy <https://pypi.python.org/pypi/lensfunpy>`_ library.
"""

from __future__ import division, absolute_import, print_function

import numpy as np
import lensfunpy
import lensfunpy.util

from auromat.util import exiftool
from auromat.util.image import loadImage, saveImage
from auromat.utils import vectorLengths

class LensNotFoundInEXIFError(Exception):
    pass
class CameraNotFoundInEXIFError(Exception):
    pass
class CameraNotFoundInDBError(Exception):
    pass
class LensNotFoundInDBError(Exception):
    pass

distortionTags = ['EXIF:Model', 'EXIF:Make', 'Composite:LensID',
                  # and some additional tags used by getLensfunModifier():
                  'EXIF:FocalLength#','Composite:Aperture#', 'Composite:ImageSize']

def _readExifTags(imagePath, tagNames, exiftoolObj=None):
    if exiftoolObj:
        assert not exiftoolObj.nums
        tags = exiftoolObj.get_tags(tagNames, imagePath)
    else:
        with exiftool.ExifTool(nums=False) as et:
            tags = et.get_tags(tagNames, imagePath)
    return tags

def findCameraAndLensFromExif(tags, lensfunDbObj=None, minAcceptedScore=100, raiseIfNotFoundInDB=True):
    if not tags.get('Composite:LensID'):
        raise LensNotFoundInEXIFError('No LensID in EXIF data')
    
    if not tags.get('EXIF:Make') or not tags.get('EXIF:Model'):
        raise CameraNotFoundInEXIFError('No camera make/model in EXIF data')
 
    if lensfunDbObj is None:
        db = lensfunpy.Database()
    else:
        db = lensfunDbObj
    
    cams = db.find_cameras(tags['EXIF:Make'], tags['EXIF:Model'], loose_search=False)
    if not cams:
        if raiseIfNotFoundInDB:
            raise CameraNotFoundInDBError('Camera "' + tags['EXIF:Make'] + ' - ' + tags['EXIF:Model'] + '" not found in DB!')
        else:
            return None, None
    # FIXME cam score is always 0
#    if cams[0].score < minAcceptedScore:
#        raise CameraNotFoundInDBError('Camera "' + tags['EXIF:Make'] + ' - ' + tags['EXIF:Model'] + '" not found! ' + 
#                                  'Closest was "' + cams[0].maker + ' - ' + cams[0].model + '" with score ' + 
#                                  str(cams[0].score) + ' (<' + str(minAcceptedScore) + ')' )
    cam = cams[0]
     
    lenses = db.find_lenses(cam, None, tags['Composite:LensID'], loose_search=True)
    if not lenses or lenses[0].score < minAcceptedScore:
        if raiseIfNotFoundInDB:
            if not lenses:
                raise LensNotFoundInDBError('Lens "' + tags['Composite:LensID'] + '" not found in DB!')
            if lenses[0].Score < minAcceptedScore:
                raise LensNotFoundInDBError('Lens "' + tags['Composite:LensID'] + '" not found in DB! ' +
                                        'Closest was "' + lenses[0].model + '" with score ' + 
                                        str(lenses[0].score) + ' (<' + str(minAcceptedScore) + ')' )
        else:
            return cam, None
        
    lens = lenses[0]
    if lens.model != tags['Composite:LensID']:
        print('NOTE: Using lensfun lens "' + lens.model + '" with score ' + str(lens.score) + \
              ' (EXIF: "' + tags['Composite:LensID'] + '")')

    return cam, lens

def findCameraAndLens(imagePath, lensfunDbObj=None, exiftoolObj=None, retTags=False, minAcceptedScore=100):
    """
    Finds camera and lens from EXIF data.
    
    Note: The lensfunDbObj reference must be kept as long as the returned camera and lens
    objects shall be used. Once lensfunDbObj gets garbage collected, the camera and lens
    objects get unusable (as their underlying C objects get freed). 
    """
    tags = _readExifTags(imagePath, distortionTags, exiftoolObj)
    cam, lens = findCameraAndLensFromExif(tags, lensfunDbObj, minAcceptedScore)
    
    if retTags:
        return cam, lens, tags
    else:
        return cam, lens

def getLensfunModifierFromExif(tags, width=None, height=None, lensfunDbObj=None, distance=10000):
    """
    WARNING: Not setting width and height may produce surprising results for RAW files.
    If width and height are not set, then Composite:ImageSize is used.
    This tag contains the full RAW size, but many RAW decoders produce slightly
    cropped images. Therefore it may be necessary to first decode the RAW image
    and determine the width and height directly.
             
    :param dict tags: must contain 'EXIF:Model', 'EXIF:Make', 'Composite:LensID',
                      'EXIF:FocalLength', 'Composite:Aperture' 
                      and optionally 'Composite:ImageSize' if width,height is None 
    """
    cam, lens = findCameraAndLensFromExif(tags, lensfunDbObj)
    
    if width is None:
        width, height = tags['Composite:ImageSize'].split('x')
        width, height = int(width), int(height)
    
    mod = lensfunpy.Modifier(lens, cam.crop_factor, width, height)
    mod.initialize(tags['EXIF:FocalLength'], tags['Composite:Aperture'], distance=distance)

    return mod, cam, lens

def getLensfunModifierFromParams(model, params, width, height):
    """
    
    :param str model: 'ptlens', 'poly3', or 'poly5'
    :param list params: a list of 1, 2 or 3 parameters, depending on the model
    :param width: image width in pixels
    :param height: image height in pixels
    """
    xml = lensfunXML(model, *params)
    db = lensfunpy.Database(xml=xml, load_common=False)
    cam = db.cameras[0]
    lens = db.lenses[0]
    mod = lensfunpy.Modifier(lens, cam.crop_factor, width, height)
    mod.initialize(1, 1)
    return mod

def getLensfunModifier(imagePath, width=None, height=None, lensfunDbObj=None, exiftoolObj=None, distance=10000):
    """
    See :func:`getLensfunModifierFromExif` for a WARNING on not setting width and height.
    """
    tags = _readExifTags(imagePath, distortionTags, exiftoolObj)
    mod, cam, lens = getLensfunModifierFromExif(tags, width, height, lensfunDbObj, distance)
    return mod, cam, lens

def correctLensDistortion(imagePath, undistImagePath, 
                          lensfunDbObj=None, mod=None, 
                          exiftoolObj=None, preserveExif=True,
                          **saveImageKws):
    """
    Correct lens distortion of an image using its EXIF headers and the lensfun library.
    If the camera or lens are not found in the lensfun database, an exception is raised.
    
    :param undistImagePath: path to output image; folders must already exist! 
    :param exiftoolObj: if not None, use the given exiftool object (must have nums=False)
    :param lensfunDbObj: if not None, use the given lensfun.Database object
    :param mod: if not None, use this Modifier instead of calling getLensfunModifier()
                Note: lensfunDbObj is not used in this case.
    :raise ValueError: when lens wasn't found in EXIF data
    :raise CameraNotFoundInDBError: when the camera wasn't found in lensfun database
    :raise LensNotFoundInDBError: when the lens wasn't found in lensfun database  
    """
    im = loadImage(imagePath)
    if mod is None:
        height, width = im.shape[0], im.shape[1]
        mod, _, _ = getLensfunModifier(imagePath, width, height, lensfunDbObj, exiftoolObj)

    undistCoords = mod.apply_geometry_distortion()    
    imUndistorted = lensfunpy.util.remap(im, undistCoords)
    saveImage(undistImagePath, imUndistorted, **saveImageKws)
    
    # TODO set a flag indicating that lens correction has been done
    #      there is no standard flag for that
    #      the closest one is Xmp.digiKam.LensCorrectionSettings
    #      see http://api.kde.org/extragear-api/graphics-apidocs/digikam/html/classDigikam_1_1LensFunFilter.html
    #      adding custom XMP tags requires to change some exiftool config file...
    if preserveExif:
        if exiftoolObj:
            exiftoolObj.copy_tags(imagePath, undistImagePath)
        else:
            with exiftool.ExifTool() as et:
                et.copy_tags(imagePath, undistImagePath)

def lensfunXML(model, *params):
    """
    Return XML in lensfun DB format for the specified distortion profile.
    """
    if model == 'ptlens':
        dist = 'model="ptlens" a="{a}" b="{b}" c="{c}"'.format(a=params[0], b=params[1], c=params[2])
    elif model == 'poly3':
        dist = 'model="poly3" k1="{k1}"'.format(k1=params[0])
    elif model == 'poly5':
        dist = 'model="poly5" k1="{k1}" k2="{k2}"'.format(k1=params[0], k2=params[1])
    else:
        raise ValueError
        
    return """
<lensdatabase>
    <mount>
        <name>Generic</name>
    </mount>
    <camera>
        <maker>CUSTOM</maker>
        <model>CUSTOM</model>
        <mount>Generic</mount>
        <cropfactor>1.0</cropfactor>
    </camera>
    <lens>
        <maker>CUSTOM</maker>
        <model>CUSTOM</model>
        <mount>Generic</mount>
        <cropfactor>1.0</cropfactor>
        <calibration>
            <distortion focal="1" {dist} />
        </calibration>
    </lens>
</lensdatabase>
    """.format(dist=dist)
             
def lensDistortionPixelDistances(imagePath=None, mod=None, lensfunDbObj=None, exiftoolObj=None, retH=False):
    """
    Return the difference between the distances from the image center to the distorted and undistorted
    pixel locations.
    
    See :func:`correctLensDistortion` for parameters.
    
    :rtype: ndarray of shape (h,w)
    """
    if not mod:
        mod,_,_ = getLensfunModifier(imagePath, lensfunDbObj, exiftoolObj)
        
    undistCoordsXY = mod.apply_geometry_distortion()

    height, width = undistCoordsXY.shape[0], undistCoordsXY.shape[1]

    y, x = np.mgrid[0:undistCoordsXY.shape[0], 0:undistCoordsXY.shape[1]]
    coordsXY = np.dstack((x,y))
    
    center = np.array([width/2, height/2])
    
    vectorsDist = (coordsXY - center).reshape(-1, 2)
    vectorsUndist = (undistCoordsXY - center).reshape(-1, 2)
    
    hDist = vectorLengths(vectorsDist).reshape(coordsXY.shape[0], coordsXY.shape[1])
    hUndist = vectorLengths(vectorsUndist).reshape(coordsXY.shape[0], coordsXY.shape[1])
    
    distance = hDist - hUndist
    
    if retH:
        return distance, hDist, hUndist
    else:
        return distance



# Copyright European Space Agency, 2013

"""
This module allows to easily download images
from `NASA's Earth Observation website <http://eol.jsc.nasa.gov>`_ in JPEG and RAW format.

As there is no API on NASA's end, we have to rely on a certain URL structure,
and, in case of RAW files, HTML structure. As this is not robust, it may fail
if the website is restructured. In that case, the code has to be adapted. 

Notes on JPEGs
--------------
JPEGs from the EOL archive are produced from the RAW camera files.
Different post-processing settings have been applied during the production
of these JPEGs, e.g. colour or exposure correction, or 180 degree rotation.
This is also true for images that belong to a single sequence of images, that is,
colour/exposure might change suddenly. The JPEGs are therefore not
suitable for scientific purposes. On the other hand, they often have hot pixels
removed. A lens distortion correction is typically not applied.

In theory the JPEGs can be used for astrometry while later using the RAW files
for scientic purposes. This is however a risky process as it has to be guaranteed
that the image orientation matches and that the lens distortion was not corrected
in the JPEGs already. To prevent checking these things each time, it is better to
use the RAW files in the first place and prepare them for astrometry ourselves,
that is, removing hot pixels and possibly noise.
"""

from __future__ import division, print_function, absolute_import

from six.moves import range
import six.moves.urllib as urllib
import os
import re
import shutil
import json
import warnings
from time import sleep
from datetime import datetime, timedelta
from collections import namedtuple

from auromat.util import exiftool
from auromat.util.os import makedirs
from auromat.util.url import urlResponseCode, downloadFiles 

try:
    import lensfunpy
    from auromat.util.lensdistortion import CameraNotFoundInDBError, LensNotFoundInDBError,\
        CameraNotFoundInEXIFError, LensNotFoundInEXIFError, getLensfunModifier
    import auromat.util.lensdistortion
except Exception as e:
    print(str(e))
    warnings.warn('lensfunpy not found, reduced functionality in auromat.solving.eol')

metadataFilename = 'meta.json'

class SequenceMetadata(object):
    def __init__(self, mission, roll, fromFrame, toFrame, pattern, frameGaps, lensDistortionCorrected,
                 lensDistortionCorrectionParams=None):
        self.mission = mission
        self.roll = roll
        self.fromFrame = fromFrame
        self.toFrame = toFrame
        self.pattern = pattern
        self.frameGaps = frameGaps
        self.lensDistortionCorrected = lensDistortionCorrected
        self.lensDistortionCorrectionParams = lensDistortionCorrectionParams

class LensDistortionCorrectionParams(object):
    def __init__(self, cameraMaker, cameraModel, cameraVariant, lensMaker, lensModel, 
                 focalLength, aperture):  
        self.cameraMaker = cameraMaker
        self.cameraModel = cameraModel
        self.cameraVariant = cameraVariant
        self.lensMaker = lensMaker
        self.lensModel = lensModel
        self.focalLength = focalLength
        self.aperture = aperture

jpgUrlPattern = 'http://eol.jsc.nasa.gov/sseop/images/ESC/large/{mission}/{mission}-{roll}-{frame}.JPG'
jpgFilePattern = '{mission}-{roll}-{frame}.jpg' # the filename on disk

# 'file' is extracted from the photoPage
photoPageUrlPattern = 'http://eol.jsc.nasa.gov/scripts/sseop/photo.pl?mission={mission}&roll={roll}&frame={frame}'
rawFilePhotoPagePattern = r'<a href=RequestOriginalImage.pl\?mission=[A-Z\d]+&roll=[A-Z\d]+&frame=[\d]+&file=([\w\.]+)>'
rawRequestUrlPattern = 'http://eol.jsc.nasa.gov/scripts/sseop/RequestOriginalImage.pl?mission={mission}&roll={roll}&frame={frame}&file={file}'
rawUrlPattern = 'http://eol.jsc.nasa.gov/sseop/OriginalImagery/{file}'
rawFilePatternNoExt = '{mission}-{roll}-{frame}' # extension not in here, could be .nef or something else

auroraVideosUrl = 'http://eol.jsc.nasa.gov/Videos/CrewEarthObservationsVideos/Videos_Aurora.htm'
auroraVideosPattern = r'<a name="([a-zA-Z\d_]+)">(.+?)</a>.+?' +\
                      '<a href="/scripts/sseop/photo.pl\?mission=([A-Z\d]+)&roll=([A-Z\d]+)&frame=([\d]+)" target="_blank">' +\
                      '<nobr>[A-Z\d-]+</a> to ' +\
                      '<a href="/scripts/sseop/photo.pl\?mission=([A-Z\d]+)&roll=([A-Z\d]+)&frame=([\d]+)" target="_blank">'

def downloadImages(folderPath, ids, format_):
    """
    Download images given by (mission,roll,frame) tuples in the specified
    format.
    
    Note: Use :func:`downloadImageSequence` to download a consecutive sequence
    of images. This function handles frame gaps (gaps in numbering) properly.
    
    :param folderPath: download location
    :param ids: list of tuples (mission,roll,frame)
    :param format_: jpg or raw
    """
    if format_ == 'jpg':
        return downloadImagesJpg(folderPath, ids)
    elif format == 'raw':
        # TODO implement RAW download of single frames
        raise NotImplementedError
    else:
        raise ValueError('Unknown format: ' + format_)

def downloadImagesJpg(folderPath, ids):
    """
    Download JPEG images given by (mission,roll,frame) tuples and return paths if successfull.
    On any error, False is returned.
    Files that are already existing are not downloaded again.
    
    :param folderPath: download location
    :param ids: list of tuples (mission,roll,frame)
    :rtype: list of str | False
    """
    urls = [jpgUrlPattern.format(mission=mission, roll=roll, frame=frame)
            for mission, roll, frame in ids]
    paths = [os.path.join(folderPath, jpgFilePattern.format(mission=mission, 
                                                            roll=roll, 
                                                            frame=frame))
             for mission, roll, frame in ids]
    
    makedirs(folderPath)
    
    if downloadFiles(urls, paths):
        return paths
    else:
        return False

def downloadImageSequence(folderPath, mission, fromFrame, toFrame, format_, roll='E', lensDistortionCorrected=False):
    """
    Download an image sequence in the specified format and return
    a tuple (metadata, []) on success or (False, errors) in case of errors.
    
    :param folderPath: download location
    :param format_: jpg or raw
    :rtype: tuple (SequenceMetadata, failure list)
    """
    if format_ == 'jpg':
        return _downloadImageSequenceJpg(folderPath, mission, fromFrame, toFrame, roll, 
                                         lensDistortionCorrected=lensDistortionCorrected)
    elif format_ == 'raw':
        return _downloadImageSequenceRaw(folderPath, mission, fromFrame, toFrame, roll)
    else:
        raise ValueError('Unsupported format: ' + format_)

def _downloadImageSequenceJpg(folderPath, mission, fromFrame, toFrame, roll='E', lensDistortionCorrected=False):
    # first, download in temp folder, then copy over and remove temp folder if successful
    tempFolderPath = os.path.join(folderPath, 'in_progress')
    
    metadataPath = os.path.join(folderPath, metadataFilename)
    
    fromFrame, toFrame = int(fromFrame), int(toFrame)
    
    # check if already fully downloaded
    firstImage = os.path.join(folderPath, jpgFilePattern.format(mission=mission, 
                                                                roll=roll, 
                                                                frame=fromFrame))
    if os.path.exists(firstImage):
        # as the files are only moved over at the very end, it is enough to
        # check for existance of the first image
        
        # write metadata if not existing yet (for whatever reason..)
        if not os.path.exists(metadataPath):
            frameGaps = []
            for frame in range(fromFrame, toFrame+1):
                imagePath = os.path.join(folderPath, jpgFilePattern.format(mission=mission, 
                                                                           roll=roll, 
                                                                           frame=frame))
                if not os.path.exists(imagePath):
                    frameGaps.append(frame)
            
            meta = SequenceMetadata(mission=mission, roll=roll, fromFrame=fromFrame,
                                    toFrame=toFrame, pattern=jpgFilePattern,
                                    frameGaps=frameGaps, 
                                    lensDistortionCorrected=lensDistortionCorrected)
            storeMetaData(metadataPath, meta)
        else:
            meta = loadMetaData(metadataPath)

        return meta, []
    
    makedirs(folderPath, tempFolderPath)    
    
    frames = range(fromFrame, toFrame+1)
    
    urls = [jpgUrlPattern.format(mission=mission, roll=roll, frame=frame)
            for frame in frames]
    paths = [os.path.join(tempFolderPath, jpgFilePattern.format(mission=mission, 
                                                                roll=roll, 
                                                                frame=frame))
             for frame in frames]
    
    print('downloading sequence frames', fromFrame, 'to', toFrame, 'of', mission + '-' + roll) 
    
    _, errors = downloadFiles(urls, paths, retFailures=True)
    
    # We ignore 404s for frames which are not the start or end.
    # This is because there are sometimes gaps in frame numbers.
    # E.g. for ISS030 the frames 115426 to 115442 don't exist within the
    #      sequence 114986 to 115574
    failures = [] 
    frameGaps = []
    for url, error in errors:
        if isinstance(error, urllib.error.HTTPError):
            i = urls.index(url)
            frame = frames[i]
            if error.code == 404:
                if fromFrame < frame < toFrame:
                    frameGaps.append(frame)
                    continue
                else:
                    raise ValueError('Start/end frame ' + str(frame) + ' not downloadable (404)')
            else:
                failures.append((url, error.code))
        else:
            failures.append((url, error))
        
    if len(failures) > 0:
        return False, failures

    for filename in os.listdir(tempFolderPath):
        shutil.move(os.path.join(tempFolderPath, filename), folderPath)
    
    os.rmdir(tempFolderPath)
    
    meta = SequenceMetadata(mission=mission, roll=roll, fromFrame=fromFrame,
                            toFrame=toFrame, pattern=jpgFilePattern,
                            frameGaps=frameGaps, 
                            lensDistortionCorrected=lensDistortionCorrected)
    storeMetaData(metadataPath, meta)
    
    return meta, []

def _downloadImageSequenceRaw(folderPath, mission, fromFrame, toFrame, roll='E'):
    assert roll == 'E' # only those have RAW files
    
    # first, download in temp folder, then copy over and remove temp folder if successful
    tempFolderPath = os.path.join(folderPath, 'in_progress')
    
    metadataPath = os.path.join(folderPath, metadataFilename)
    
    fromFrame, toFrame = int(fromFrame), int(toFrame)
    
    # check if already fully downloaded
    if os.path.exists(metadataPath):
        return True
    
    makedirs(folderPath, tempFolderPath)

    # first, we determine the RAW filename pattern by looking at a photo page
    firstPhotoPageUrl = photoPageUrlPattern.format(mission=mission, roll=roll, frame=fromFrame)
    photoPageContent = urllib.request.urlopen(firstPhotoPageUrl).read()
    
    match = re.search(rawFilePhotoPagePattern, photoPageContent)
    if match is None:
        raise RuntimeError('Could not find RAW filename on page ' + firstPhotoPageUrl)
    
    rawFilename = match.group(1)
    rawFileBase, rawFileExt = os.path.splitext(rawFilename)
    
    assert mission in rawFileBase or mission.lower() in rawFileBase
    assert roll in rawFileBase or roll.lower() in rawFileBase
    assert str(fromFrame) in rawFileBase
    
    rawFileBasePattern = rawFileBase
    if mission in rawFileBase:
        rawFileBasePattern = rawFileBase.replace(mission, '{mission}')
        missionCased = mission
    elif mission.lower() in rawFileBasePattern:
        rawFileBasePattern = rawFileBasePattern.replace(mission.lower(), '{mission}')
        missionCased = mission.lower()
    else:
        raise RuntimeError('Could not find mission name in ' + rawFileBase)

    if roll in rawFileBasePattern:
        rawFileBasePattern = rawFileBasePattern.replace(roll, '{roll}')
        rollCased = roll
    elif roll.lower() in rawFileBasePattern:
        rawFileBasePattern = rawFileBasePattern.replace(roll.lower(), '{roll}')
        rollCased = roll.lower()
    else:
        raise RuntimeError('Could not find roll name in ' + rawFileBase)
    
    frameZfilled = lambda frame: str(frame).zfill(6)
    if frameZfilled(fromFrame) in rawFileBasePattern:
        rawFileBasePattern = rawFileBasePattern.replace(frameZfilled(fromFrame), '{frame}')
        frameFn = frameZfilled
    elif str(fromFrame) in rawFileBasePattern:
        rawFileBasePattern = rawFileBasePattern.replace(str(fromFrame), '{frame}')
        frameFn = str
    else:
        raise RuntimeError('Could not find frame number in ' + rawFileBase)

    rawFilenamePattern = rawFileBasePattern + rawFileExt
    print('Raw filename pattern: ' + rawFilenamePattern)
    
    frames = range(fromFrame, toFrame+1)
    rawFilenames = [rawFilenamePattern.format(mission=missionCased, roll=rollCased, frame=frameFn(frame))
                    for frame in frames]
    rawRequestUrls = [rawRequestUrlPattern.format(mission=mission, roll=roll, frame=frame, file=rawFilename)
                      for frame, rawFilename in zip(frames, rawFilenames)]
    rawUrls = [rawUrlPattern.format(file=rawFilename) for rawFilename in rawFilenames]
    
    rawFilePatternDisk = rawFilePatternNoExt + rawFileExt.lower()
    rawFilenamesDisk = [rawFilePatternDisk.format(mission=mission, roll=roll, frame=frame) for frame in frames]
    paths = [os.path.join(tempFolderPath, rawFilenameDisk) for rawFilenameDisk in rawFilenamesDisk]
    
    # jpg URLs are used to check if the frame exists (or whether there's a frame gap)
    jpgUrls = [jpgUrlPattern.format(mission=mission, roll=roll, frame=frame)
               for frame in frames]
    
    frameGaps = []
    failures = []
    queue = []
    
    for frame, jpgUrl, rawUrl, rawRequestUrl, path in zip(frames, jpgUrls, rawUrls, rawRequestUrls, paths):
        if os.path.exists(path):
            continue
        try:
            code = urlResponseCode(jpgUrl)
            if code == 200:
                queue.append((rawUrl, rawRequestUrl, path))
                print('Got 200, added frame ' + str(frame) + ' to queue')
            elif code == 404:
                if fromFrame < frame < toFrame:
                    frameGaps.append(frame)
                    print('Got 404, ignoring frame ' + str(frame))
                else:
                    raise ValueError('Start/end frame ' + str(frame) + ' not downloadable (404)')
            else:
                failures.append((rawUrl, code))
                print('Failure: Unexpected response code for jpgUrl: ' + str(code))
        except Exception as e:
            failures.append((rawUrl, e))
            print('Failure: ' + repr(e))
            
    # download RAW files in batches to avoid overloading the server
    batchSize = 30
    batches = [queue[i:i+batchSize] for i in range(0, len(queue), batchSize)]
    for batch in batches:
        batchUrls = []
        batchPaths = []
        for rawUrl, rawRequestUrl, path in batch:
            try:
                code = urlResponseCode(rawRequestUrl)
                if code == 200:
                    print('queried ' + rawRequestUrl)
                    batchUrls.append(rawUrl)
                    batchPaths.append(path)
                else:
                    failures.append((rawUrl, code))
                    print('Failure: Unexpected response code for rawRequestUrl: ' + str(code))
            except Exception as e:
                failures.append((rawUrl, e))
                print('Failure: ' + repr(e))
        
        # now check rawUrls until the files are available for download
        # The request "may take 5 minutes or more to complete" (quote from request page)
        success, failures_ = downloadFiles(batchUrls, batchPaths, retFailures=True)
        failureCount = len(failures_)
        lastFailureCountDecrease = datetime.now()
        while not success and datetime.now() - lastFailureCountDecrease < timedelta(minutes=8):
            sleep(30)
            success, failures_ = downloadFiles(batchUrls, batchPaths, retFailures=True)
            if len(failures_) < failureCount:
                lastFailureCountDecrease = datetime.now()
                failureCount = len(failures_)
        
        failures.extend(failures_)
        
    if len(failures) > 0:
        return False, failures
        
    for filename in os.listdir(tempFolderPath):
        shutil.move(os.path.join(tempFolderPath, filename), folderPath)
    
    os.rmdir(tempFolderPath)
    
    meta = SequenceMetadata(mission=mission, roll=roll, fromFrame=fromFrame,
                            toFrame=toFrame, pattern=rawFilePatternDisk,
                            frameGaps=frameGaps,
                            lensDistortionCorrected=False)
    storeMetaData(metadataPath, meta)
    
    return meta, [] 

Sequence = namedtuple('Sequence', ['mission', 'roll', 'fromFrame', 'toFrame', 'title', 'urlAnchor'])
def extractAuroraSequences():
    """
    Extracts metadata of all sequences found on
    http://eol.jsc.nasa.gov/Videos/CrewEarthObservationsVideos/Videos_Aurora.htm.
    """
    content = urllib.request.urlopen(auroraVideosUrl).read()
    sequences = []
    for match in re.finditer(auroraVideosPattern, content, re.DOTALL):
        urlAnchor, title = match.group(1,2)
        mission, roll, fromFrame = match.group(3,4,5)
        mission_, roll_, toFrame = match.group(6,7,8)
        assert mission == mission_ and roll == roll_
        sequences.append(Sequence(mission, roll, int(fromFrame), int(toFrame), title, urlAnchor))
        
    return sequences

def storeMetaData(jsonPath, meta):
    if os.path.isdir(jsonPath):
        jsonPath = os.path.join(jsonPath, metadataFilename)
    
    metaDict = meta.__dict__
    if meta.lensDistortionCorrectionParams is not None:
        metaDict['lensDistortionCorrectionParams'] = meta.lensDistortionCorrectionParams.__dict__
    
    with open(jsonPath, 'w') as fp:
        json.dump(metaDict, fp, indent=4)

def loadMetaData(jsonPath):
    if os.path.isdir(jsonPath):
        jsonPath = os.path.join(jsonPath, metadataFilename)
    print('loading ' + jsonPath)
    with open(jsonPath) as fp:
        metaDict = json.load(fp)
    meta = SequenceMetadata(**metaDict)
    if meta.lensDistortionCorrectionParams is not None:
        meta.lensDistortionCorrectionParams = LensDistortionCorrectionParams(**meta.lensDistortionCorrectionParams)
    return meta

def filenameOf(frame, meta):
    return _filenameOf(meta.mission, meta.roll, frame, meta.pattern)
    
def _filenameOf(mission, roll, frame, pattern):
    return pattern.format(mission=mission,
                          roll=roll,
                          frame=frame)

def frameIter(meta):
    for frame in range(meta.fromFrame, meta.toFrame+1):
        if frame not in meta.frameGaps:
            yield frame
            
def filenameIter(meta):
    for frame in frameIter(meta):
        yield filenameOf(frame, meta), frame

def correctLensDistortion(folderPath, undistFolderPath, lensfunDbObj=None):
    """
    Corrects the lens distortion of all images in `folderPath` using
    lensfun's distortion profile database.
    
    It is assumed that all images have the same camera and lens.
    Images are skipped whose corrected version already exists in undistFolderPath.
    """
    meta = loadMetaData(folderPath)
    firstImagePath = os.path.join(folderPath, filenameOf(meta.fromFrame, meta))
    mod, cam, lens = getLensfunModifier(firstImagePath, lensfunDbObj=lensfunDbObj)
    
    makedirs(undistFolderPath)
    
    print('starting lens distortion correction for ' + folderPath)
    
    splitted = os.path.splitext(meta.pattern)
    undistPattern = splitted[0] + '_dc' + splitted[1]
       
    with exiftool.ExifTool() as et:
        for filename, frame in filenameIter(meta):
            imagePath = os.path.join(folderPath, filename)
            filenameUndist = _filenameOf(meta.mission, meta.roll, frame, undistPattern)
            undistImagePath = os.path.join(undistFolderPath, filenameUndist)
            if os.path.exists(undistImagePath):
                continue
            auromat.util.lensdistortion.correctLensDistortion(imagePath, undistImagePath, exiftoolObj=et, mod=mod)
    
    dcParams = LensDistortionCorrectionParams(cam.maker, cam.model, cam.variant, lens.maker, lens.model,
                                              mod.focal_length, mod.aperture)
    
    meta.pattern = undistPattern
    meta.lensDistortionCorrected = True
    meta.lensDistortionCorrectionParams = dcParams
    
    storeMetaData(undistFolderPath, meta)
    
if __name__ == '__main__':
    from os.path import expanduser
    home = expanduser("~")
    rootFolderPath = os.path.join(home, 'data/images/aurora_sequences_jpg')
    rootUndistFolderPath = os.path.join(home, 'data/images/aurora_sequences_jpg_dc')
    success, _ = downloadAuroraSequences(rootFolderPath, 'jpg')
    if success:
        print('All aurora sequences were downloaded successfully')
                
        lensfunDb = lensfunpy.Database()
        
        folders = os.listdir(rootFolderPath)
        folderPaths = [os.path.join(rootFolderPath, f) for f in folders]
        folderPaths = filter(os.path.isdir, folderPaths)
        
        for folderPath in folderPaths:
            undistFolderPath = os.path.join(rootUndistFolderPath, os.path.basename(folderPath))
            try:
                correctLensDistortion(folderPath, undistFolderPath, lensfunDb)
            except (CameraNotFoundInEXIFError, LensNotFoundInEXIFError, LensNotFoundInDBError, CameraNotFoundInDBError) as e:
                print('Lens correction not done for ' + folderPath + ': ' + str(e))

    


# Copyright European Space Agency, 2013

import unittest
from auromat.util.movie import createMovie
import os
import tempfile


class Test(unittest.TestCase):

    def testMp4Movie(self):
        imagePaths = [getResourcePath('ISS030-E-102170_dc.jpg')]*30
        moviePath = tempfile.mktemp(suffix='.mp4')
        _createTempMovie(moviePath, lambda: createMovie(moviePath, imagePaths, width=1280))
            
    def testWebMMovie(self):
        imagePaths = [getResourcePath('ISS030-E-102170_dc.jpg')]*30
        moviePath = tempfile.mktemp(suffix='.webm')
        _createTempMovie(moviePath, lambda: createMovie(moviePath, imagePaths, width=1280))
            
def _createTempMovie(moviePath, fn):
    try:
        fn()
    except:
        # if there was an exception we want to see it, and
        # ignore the one we might get from removing a nonexisting file
        if os.path.exists(moviePath):
            os.remove(moviePath)
        raise
    else:
        os.remove(moviePath)
    
def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)


# Copyright European Space Agency, 2013

from __future__ import division, print_function

import os
import collections

import numpy as np
import numpy.ma as ma

from spacepy import pycdf

from auromat.mapping.mapping import BaseMappingProvider, \
    sanitize_data, BaseMapping, ArrayImageMixin, GenericMapping
from auromat.util.decorators import inherit_docs
import auromat.utils

@inherit_docs
class CDFMappingProvider(BaseMappingProvider):
    def __init__(self, cdfPaths, maxTimeOffset=3):
        self.cdfPaths = cdfPaths
        self.maxTimeOffset = maxTimeOffset
        
        # create mapping from date to (path index, cdf index)
        # cdf index is the index inside the cdf file
        # currently we only store 1 image in each, but this could change
        # e.g. storing multiple downsampled images in one file
        
        datemap = {}
        for path_idx, path in enumerate(cdfPaths):
            with pycdf.CDF(path) as root:
                dates = root['Epoch'][:]
                for cdf_idx, date in enumerate(dates):
                    if date in datemap:
                        raise ValueError('The date ' + str(date) + ' is appearing twice ' +
                                         'in the CDF files ' + path + ' and ' +
                                         cdfPaths[datemap[date][0]])
                    datemap[date] = (path_idx, cdf_idx)
        
        self.datemap = collections.OrderedDict(sorted(datemap.items()))
                     
    def __len__(self):
        return len(self.datemap)

    @property
    def range(self):
        return list(self.datemap.keys())[0], list(self.datemap.keys())[-1]
    
    def contains(self, date):
        dates = list(self.datemap.keys())
        idx = auromat.utils.findNearest(dates, date)
        offset = abs(dates[idx]-date).total_seconds()
        return offset <= self.maxTimeOffset
        
    def get(self, date):
        dates = list(self.datemap.keys())
        idx = auromat.utils.findNearest(dates, date)
        offset = abs(dates[idx]-date).total_seconds()
        if offset > self.maxTimeOffset:
            raise ValueError('Closest mapping found at ' + str(dates[idx]) + 
                             ' but offset > ' + str(self.maxTimeOffset) + ' seconds, ' +
                             'requested: ' + str(date))
        path_idx, cdf_idx = self.datemap[dates[idx]]
        return CDFMapping(self.cdfPaths[path_idx], cdf_idx)
        
    def getById(self, identifier):
        raise NotImplementedError
    
    def getSequence(self, dateBegin=None, dateEnd=None):
        if not dateBegin:
            dateBegin = self.range[0]
        if not dateEnd:
            dateEnd = self.range[1]
        dates = filter(lambda d: dateBegin <= d <= dateEnd, self.datemap.keys())
        for date in dates:
            path_idx, cdf_idx = self.datemap[date]
            yield CDFMapping(self.cdfPaths[path_idx], cdf_idx)

@sanitize_data
@inherit_docs
class CDFMapping(ArrayImageMixin, BaseMapping):
    
    # variable names in the cdf files as produced by auromat.export.cdf
    var_altitude = 'altitude'
    var_cameraPos = 'camera_pos'
    var_photoTime = 'Epoch'
    var_img = 'img'
    var_img_red = 'img_red'
    var_img_green = 'img_green'
    var_img_blue = 'img_blue'
    var_latsCenter = 'lat'
    var_lonsCenter = 'lon'
    var_zenithAngle = 'zenith_angle'
    
    def __init__(self, cdfPath, i=0):
        with pycdf.CDF(cdfPath) as root:
            var = root
            altitude = var[self.var_altitude][...]/1000
            cameraPosGCRS = var[self.var_cameraPos][i]
            photoTime = var[self.var_photoTime][i]
            
            # for three channels (RGB), each channel is stored as a
            # separate variable: img_red, img_green, img_blue
            # for grayscale, the single variable is called 'img'
            try:
                fillval = var[self.var_img].attrs['FILLVAL']
                img = np.atleast_3d(var[self.var_img][i])
                img = _convertImgDtype(img, fillval)
            except:
                fillval = var[self.var_img_red].attrs['FILLVAL']
                img_red = _convertImgDtype(var[self.var_img_red][i], fillval)
                img_green = _convertImgDtype(var[self.var_img_green][i], fillval)
                img_blue = _convertImgDtype(var[self.var_img_blue][i], fillval)
                img = ma.dstack((img_red, img_green, img_blue))
                        
            latsCenter = var[self.var_latsCenter][i]
            lonsCenter = var[self.var_lonsCenter][i]
            lats = var[var[self.var_latsCenter].attrs['bounds']][i]
            lons = var[var[self.var_lonsCenter].attrs['bounds']][i]
            
            # TODO read in MLat/MLT as well if available
                        
            self._latsCenter = ma.masked_invalid(latsCenter)
            self._lonsCenter = ma.masked_invalid(lonsCenter)
            self._lats = ma.masked_invalid(lats)
            self._lons = ma.masked_invalid(lons)
            self._elevation = ma.masked_invalid(90 - var[self.var_zenithAngle][i])
            
            metadata = root.attrs
            
            assert var[self.var_altitude].attrs['UNITS'] == 'meters'
            assert var[self.var_cameraPos].attrs['UNITS'] == 'kilometers'
        
        identifier = os.path.splitext(os.path.basename(cdfPath))[0]
        BaseMapping.__init__(self, altitude, cameraPosGCRS, photoTime, identifier, metadata=metadata)
        ArrayImageMixin.__init__(self, img)

    @property
    def lats(self):
        return self._lats
    
    @property
    def lons(self):
        return self._lons
    
    @property
    def latsCenter(self):
        return self._latsCenter
    
    @property
    def lonsCenter(self):
        return self._lonsCenter
        
    @property
    def elevation(self):
        return self._elevation
    
    def createResampled(self, lats, lons, latsCenter, lonsCenter, elevation, img):
        mapping = GenericMapping(lats, lons, latsCenter, lonsCenter, elevation, self.altitude, img, 
                                 self.cameraPosGCRS, self.photoTime, self.identifier,
                                 metadata=self.metadata)
        return mapping
    
def _convertImgDtype(arr, fillval):
    if arr.dtype in [np.uint8, np.uint16, np.uint32]:
        return arr
    
    arr = ma.masked_equal(arr, fillval, copy=False)
    
    if arr.dtype == np.int16:
        assert 0 <= np.min(arr) <= np.max(arr) <= np.iinfo(np.uint8).max
        return arr.astype(np.uint8)
    
    elif arr.dtype == np.int32:
        assert 0 <= np.min(arr) <= np.max(arr) <= np.iinfo(np.uint16).max
        return arr.astype(np.uint16)
    
    elif arr.dtype == np.int64:
        assert 0 <= np.min(arr) <= np.max(arr) <= np.iinfo(np.uint32).max
        return arr.astype(np.uint32)
        
    else:
        raise NotImplementedError('Data type not supported: ' + str(arr.dtype))
    


# Copyright European Space Agency, 2013

import matplotlib as mpl
mpl.use('agg')
import matplotlib.pyplot as plt

# see https://github.com/matplotlib/matplotlib/issues/3197

# mem leak only occurs when backend is agg, rasterized=True, 
# and when savefig is called with SVG! for png there's no leak

for i in range(10000):
    fig, ax = plt.subplots()            

    circle1=plt.Circle((0,0),2,color='r',rasterized=True)
    ax.add_artist(circle1)

    fig.savefig('test.svg')
    plt.close(fig)
    


# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

from six.moves.urllib.error import HTTPError
from datetime import datetime, timedelta
import os

import numpy as np
import numpy.ma as ma
from spacepy import pycdf

from auromat.mapping.mapping import BaseMapping, MappingCollection,\
    BaseMappingProvider, sanitize_data
import auromat.utils
from auromat.util.decorators import inherit_docs
from auromat.coordinates.geodesic import wgs84A, wgs84B
from auromat.coordinates.transform import geodetic2Ecef, geodetic2EcefZero,\
    ecef2Geodetic, latLonToJ2000
from auromat.coordinates.intersection import ellipsoidLineIntersection
from auromat.util.os import touch
from auromat.util.url import downloadFile

stations = ['atha', 'chbg', 'ekat', 'fsim', 'fsmi', 'fykn',
            'gako', 'gbay', 'gill', 'inuv', 'kapu', 'kian',
            'kuuj', 'mcgr', 'nrsq', 'pgeo', 'pina', 'rank',
            'snap', 'snkq', 'talo', 'tpas', 'whit', 'yknf']

L1BaseUrl = 'http://themis.ssl.berkeley.edu/data/themis/thg/l1/asi/'
L2BaseUrl = 'http://themis.ssl.berkeley.edu/data/themis/thg/l2/asi/cal/'

L1Prefix = '{station}/{year}/{month}/'
L1Filename = 'thg_l1_asf_{station}_{date}_v01.cdf'
L2Filename = 'thg_l2_asc_{station}_19700101_v01.cdf'

@inherit_docs
class ThemisMappingProvider(BaseMappingProvider):
    
    def __init__(self, cdfL1CacheFolder, cdfL2CacheFolder, altitude=110, 
                 minBrightness=None, maxBrightness=None,
                 offline=False):
        BaseMappingProvider.__init__(self, maxTimeOffset=2)
        self.offline = offline
        if not os.path.exists(cdfL1CacheFolder) and not offline:
            os.makedirs(cdfL1CacheFolder)
        if not os.path.exists(cdfL2CacheFolder) and not offline:
            os.makedirs(cdfL2CacheFolder)
        self.cdfL1CacheFolder = cdfL1CacheFolder
        self.cdfL2CacheFolder = cdfL2CacheFolder
        self.altitude = altitude
        self.minBrightness = minBrightness
        self.maxBrightness = maxBrightness
        
    @property
    def range(self):
        raise NotImplementedError
        
    def contains(self, date):
        for station in stations:
            if getL1Data(self.cdfL1CacheFolder, station, date, maxTimeOffset=self.maxTimeOffset):
                return True
        return False
    
    def download(self, dateBegin, dateEnd):
        """
        Download data in the given interval.
        An error is raised if self.offline is True and the data has not
        been downloaded yet.
        """
        if not (dateBegin and dateEnd):
            raise ValueError('start and end dates must be given')
        if dateBegin > dateEnd:
            raise ValueError('start date must be earlier than end date')
        
        # cache everything between the given date range
        # split range into 1h resolution dates and download all
        dateBegin = datetime(*dateBegin.timetuple()[:4])
        dateEnd = datetime(*dateEnd.timetuple()[:4])
        hours = int((dateEnd-dateBegin).total_seconds())//60//60
        dates = [dateBegin + timedelta(hours=h) for h in range(hours+1)]
        
        for station in stations:
            if self.offline:
                if not hasL2Data(cdfL2CacheFolder, station):
                    raise RuntimeError('offline=True but L2 data not cached yet')            
            else:
                downloadL2Data(self.cdfL2CacheFolder, station)
            
            for date in dates:
                if self.offline:
                    if hasL1Data(cdfL1CacheFolder, station, date) is False:
                        raise RuntimeError('offline=True but L1 data not cached yet')
                else:
                    downloadL1Data(self.cdfL1CacheFolder, station, date)
    
    def get(self, date):
        mappings = getMappings(date, self.cdfL1CacheFolder, self.cdfL2CacheFolder, 
                               self.altitude, self.maxTimeOffset, self.minBrightness, self.maxBrightness,
                               offline=self.offline)
        if mappings.empty:
            raise ValueError('No THEMIS mappings found at ' + str(date) + ' +- ' + str(self.maxTimeOffset) + 's')
        return mappings
    
    def getById(self, identifier):
        raise NotImplementedError
    
    def getSequence(self, dateBegin, dateEnd):
        raise NotImplementedError
    
@sanitize_data
@inherit_docs
class ThemisMapping(BaseMapping):
    def __init__(self, lats, lons, latsCenter, lonsCenter, elev, alti, img, cameraPosGCRS, photoTime, 
                 station, minBrightness=None, maxBrightness=None):
        """
        
        :param lats: (h+1,w+1) in degrees
        :param lons: (h+1,w+1) in degrees
        :param latsCenter: (h,w) in degrees
        :param lonsCenter: (h,w) in degrees
        :param elev: elevation in degrees for each pixel center (h,w), can be None  
        :param alti: the altitude in km onto which the image was mapped (e.g. 110)
        :param img: masked array, (h,w) or (h,w,3) in [0,255] as int or float, can have NaN's
        :param cameraPosGCRS: [x,y,z] in km
        :param photoTime: datetime object
        :param string station: 
        :param minBrightness: 
        :param maxBrightness:
        """
        assert img.ndim == 2
        h, w = img.shape[0], img.shape[1]
        assert lats.shape == lons.shape == (h+1, w+1)
        assert elev is None or elev.shape == (img.shape[0], img.shape[1])
        
        # adapted from web filenames: RANK.2013.09.26.05.03.gif
        identifier = station + '.' + photoTime.strftime('%Y.%m.%d.%H.%M.%S')
        
        BaseMapping.__init__(self, alti, cameraPosGCRS, photoTime, identifier)
        self._img = img[:,:,None]
        self._lats = ma.masked_invalid(lats, copy=False)
        self._lons = ma.masked_invalid(lons, copy=False)
        self._latsCenter = ma.masked_invalid(latsCenter, copy=False)
        self._lonsCenter = ma.masked_invalid(lonsCenter, copy=False)
        self._elevation = ma.masked_invalid(elev, copy=False)
        # allow to be changed from outside:
        self.minBrightness = minBrightness
        self.maxBrightness = maxBrightness
    
    @property
    def lats(self):
        return self._lats
    
    @property
    def lons(self):
        return self._lons
    
    @property
    def latsCenter(self):
        return self._latsCenter
    
    @property
    def lonsCenter(self):
        return self._lonsCenter
    
    @property
    def elevation(self):
        return self._elevation
    
    @property
    def img(self):
        return self._img
    
    @property
    def img_unmasked(self):
        return self._img.data
    
    @property
    def rgb(self):
        img = self.brightness_scaled(self.img)    
        rgb = np.repeat(img, 3, 2)
        rgb = np.require(rgb, dtype=np.uint8)
        return rgb
    
    @property
    def rgb_unmasked(self):
        img = self.brightness_scaled(self.img_unmasked)
        rgb = np.repeat(img, 3, 2)
        rgb = np.require(rgb, dtype=np.uint8)
        return rgb
        
    def brightness_scaled(self, img):
        # brightness scaling, taken from thm_asi_create_mosaic.pro
        if self.minBrightness is not None or self.maxBrightness is not None:
            img = bytscl(self.img, min_=self.minBrightness, max_=self.maxBrightness, top=255)
        else:
            med = np.median(self.img[self.img > 1])
            img = self.img/med*64
            img = np.minimum(img, 255)
        return img
             
        
    def createResampled(self, lats, lons, latsCenter, lonsCenter, elevation, img):
        mapping = ThemisMapping(lats, lons, latsCenter, lonsCenter, elevation, self.altitude, img, 
                                self.cameraPosGCRS, self.photoTime, 
                                self.minBrightness, self.maxBrightness)
        return mapping
    
def bytscl(array, max_=None , min_=None , top=255 ):
    """
    :see: https://github.com/mperrin/misc_astro/blob/master/idlastro_ports/idlbase.py
    :license: https://github.com/mperrin/misc_astro/blob/master/LICENSE.rst
    """
    # see http://star.pst.qub.ac.uk/idl/BYTSCL.html
    # note that IDL uses slightly different formulae for bytscaling floats and ints. 
    # here we apply only the FLOAT formula...

    if max_ is None: max_ = np.nanmax(array)
    if min_ is None: min_ = np.nanmin(array)
    
    return np.maximum(np.minimum(  
        ((top+0.9999)*(array-min_)/(max_-min_)).astype(np.int16)
        , top),0)
        
def reproject(latLonASI, latsRef, lonsRef, heightRef, heightNew):
    """
    
    :param latLonASI: tuple of latitude,longitude of ASI
    :param latsRef: latitudes of pixel corners for reference height in degrees
    :param lonsRef: longitudes of pixel corners for reference height in degrees
    :param heightRef: reference height in km (above ground)
    :param heightNew: new height in km (above ground)
    :rtype: tuple of reprojected latitudes and longitudes 
    """
    latASI, lonASI = latLonASI
    xASI,yASI,zASI = geodetic2EcefZero(np.deg2rad(latASI), np.deg2rad(lonASI))
    
    # reconstruct direction vector as seen from ASI
    x,y,z = geodetic2Ecef(np.deg2rad(latsRef), np.deg2rad(lonsRef), heightRef)
    x -= xASI
    y -= yASI
    z -= zASI
    
    # reproject
    a = wgs84A + heightNew
    b = wgs84B + heightNew
    direction = np.transpose([x,y,z])
    intersection = ellipsoidLineIntersection(a, b, [xASI,yASI,zASI], direction.reshape(-1,3))
    intersection = intersection.reshape(direction.shape)
    x, y, z = intersection.transpose()
    latsNew, lonsNew = ecef2Geodetic(x, y, z, wgs84A, wgs84B)
    np.rad2deg(latsNew, latsNew)
    np.rad2deg(lonsNew, lonsNew)
    return latsNew, lonsNew

def maskByL2(mask, img):
    """
    Masks image pixels using the L2 mask.
    
    WARNING: Don't use this method as the L2 masks seem to contain
             inconsistent data (interpretation of 0 and 1 mixed up in one case)
    
    :param mask: L2 mask of shape (w,w)
    :param ndarray img: grayscale img (w,w)
    :rtype: ndarray<float32> with masked pixels set to NaN
    """
    img = img.astype(np.float32)
    img[mask==1] = np.nan
    
    return img

def hasL2Data(cdfL2CacheFolder, station):
    filename = L2Filename.format(station=station)
    path = os.path.join(cdfL2CacheFolder, filename)
    return os.path.exists(path)

def downloadL2Data(cdfL2CacheFolder, station):   
    if hasL2Data(cdfL2CacheFolder, station):
        return
    
    # download remotely, if available
    filename = L2Filename.format(station=station)
    path = os.path.join(cdfL2CacheFolder, filename)
    url = L2BaseUrl + filename
    downloadFile(url, path)

def getL2Data(cdfL2CacheFolder, station):
    """
    
    :param cdfL2CacheFolder: full path to folder where L2 CDF files are/should be stored
    :param station: name of the station in lower case
    :rtype: tuple (latASI,lonASI), az, el, latsRef, lonsRef, heightsRef (km)
    :raises: DownloadError (in case of network/IO problems)
    """    
    filename = L2Filename.format(station=station)
    path = os.path.join(cdfL2CacheFolder, filename)
        
    with pycdf.CDF(path) as cdf:
        latASI = cdf['thg_asc_' + station + '_glat'][...] # station coords
        lonASI = cdf['thg_asc_' + station + '_glon'][...]
        
        # coordinates are only defined for useful pixels, rest is nan
        az = cdf['thg_asf_' + station + '_azim'][0] # azimuth of pixel center in degrees (256,256)
        el = cdf['thg_asf_' + station + '_elev'][0] # elevation of pixel center in degrees (256,256)
        
        latsRef = cdf['thg_asf_' + station + '_glat'][0] # lats in degrees of pixel corners for multiple aurora heights (257,257,3)
        lonsRef = cdf['thg_asf_' + station + '_glon'][0] # lons, dito
        
        heightsRef = cdf['thg_asf_' + station + '_alti'][...] # reference heights in meters
        
        # the masks don't contain useful and consistent data, ignore for now
        #mask = cdf['thg_asf_' + station + '_mask'][0] # mask (0 or 1) for every pixel (256,256)
        
        # multiply and flat multiply are 1.0, offset is 2500 for all stations for every pixel
#        multiply = cdf['thg_asf_' + station + '_multiply'][...]
#        flat = cdf['thg_asf_' + station + '_flat'][...]
#        offset = cdf['thg_asf_' + station + '_offset'][...]

    
    # turn (257,257,3) into (3,257,257) to make it more convenient
    latsRef = np.rollaxis(latsRef, 2)
    lonsRef = np.rollaxis(lonsRef, 2)
    
    return (latASI,lonASI), az, el, latsRef, lonsRef, heightsRef/1000

def getL1Filename(station, date):
    return L1Filename.format(station=station, date=date.strftime('%Y%m%d%H'))

def hasL1Data(cdfL1CacheFolder, station, date, retry404After = timedelta(days=30)):
    filename = getL1Filename(station, date)
    path = os.path.join(cdfL1CacheFolder, filename)
    if os.path.exists(path):
        return True
    
    path404 = path + '.404'
    if os.path.exists(path404):
        mtime = datetime.fromtimestamp(os.path.getmtime(path404))
        if datetime.now() - mtime > retry404After:
            os.remove(path404)
        else:
            return '404'
    return False

def downloadL1Data(cdfL1CacheFolder, station, date, retry404After = timedelta(days=30)):
    """
    
    :param datetime.timedelta retry404After: amount of time after which 404'd requests are retried
    """
    # check if it exists in the cache
    status = hasL1Data(cdfL1CacheFolder, station, date)
    if status is True:
        return True
    if status == '404':
        return False
    # if false, try downloading
    
    filename = getL1Filename(station, date)
    path = os.path.join(cdfL1CacheFolder, filename)
         
    # download remotely, if available
    url = L1BaseUrl + L1Prefix.format(station=station, year=date.strftime('%Y'),
                                      month=date.strftime('%m')) + filename
    
    path404 = path + '.404'
    try:
        downloadFile(url, path, unifyErrors=False)
        
    except HTTPError as e:
        if e.code == 404:
            touch(path404)
        return False
            
    except Exception as e: # e.g. no disk space left
        print(repr(e))
        return False
    
    return True

def getL1Data(cdfL1CacheFolder, station, date, maxTimeOffset=2):
    """
    Loads a single image (if available) corresponding to the given date.
    
    :param cdfL1CacheFolder:
    :param station:
    :param date:
    :param maxTimeOffset: maximum difference in seconds between nearest image and given date
    """    
    filename = getL1Filename(station, date)
    path = os.path.join(cdfL1CacheFolder, filename)
    
    with pycdf.CDF(path) as cdf:
        # search for image at given time
        epoch = cdf['thg_asf_' + station + '_epoch']
        idx = auromat.utils.findNearest(epoch, date)
        offset = abs(epoch[idx]-date).total_seconds()
        if offset > maxTimeOffset:
            print('image time offset too big:', offset, 'seconds')
            return None, None
        return cdf['thg_asf_' + station][idx], epoch[idx]

def mappingSingleASI(station, date, cdfL1CacheFolder, cdfL2CacheFolder, 
                     maxTimeOffset=2, altitude=110,
                     minBrightness=None, maxBrightness=None, offline=False):
    if offline and hasL1Data(cdfL1CacheFolder, station, date) is False:
        raise RuntimeError('offline=True but L1 data not cached yet')
    
    if not downloadL1Data(cdfL1CacheFolder, station, date):
        return None
    
    img, imgDate = getL1Data(cdfL1CacheFolder, station, date, maxTimeOffset)
    if img is None:
        return None
    
    img = ma.masked_array(img)
    
    if not offline:
        downloadL2Data(cdfL2CacheFolder, station)
    latLonASI, _, el, latsRef, lonsRef, heightsRef = getL2Data(cdfL2CacheFolder, station) 
    
    if altitude*1000 in heightsRef:
        refIdx = np.where(heightsRef==altitude*1000)[0][0]
        lats, lons = latsRef[refIdx], lonsRef[refIdx]
    else:
        # TODO reprojection leads to border artifacts (long polygons) for minElevationToPlot <= 3
        refIdx = 0
        lats, lons = reproject(latLonASI, latsRef[refIdx], lonsRef[refIdx], heightsRef[refIdx], altitude)
    
    # THEMIS mappings don't span across the discontinuity
    # so it is easy to calculate the pixel centers
    latsCenter = (lats[:-1,:-1] + lats[1:,:-1] + lats[:-1,1:] + lats[1:,1:])/4
    lonsCenter = (lons[:-1,:-1] + lons[1:,:-1] + lons[:-1,1:] + lons[1:,1:])/4
    
#    img = maskByL2(mask, img)
        
    # for THEMIS, downsampling is only useful if the resolution should be *much* lower
    # otherwise downsampling with similar pixel density just introduces additional visible artifacts 
    #lats, lons, img = auromat.resample.resample(lats, lons, img, skipDiscontinuityCheck=True, skipPoleCheck=True)
    
    # 2500 is the _offset value which is equal in all THEMIS L2 files for all pixels.
    # The multipliers _multiply and _flat are all 1.0 still. Should this change in the future,
    # then it has to be changed here as well (by properly applying the matrices). For now,
    # this is a fast short-cut.
    img -= 2500
    
    latASI, lonASI = latLonASI    
    cameraPosGCRS = latLonToJ2000(latASI, lonASI, 0, imgDate)
        
    mapping = ThemisMapping(lats, lons, latsCenter, lonsCenter, el, altitude, img, cameraPosGCRS, 
                            imgDate, station, minBrightness, maxBrightness)
    
    # THEMIS L2 data seems to be partly wrong at very low elevation angles.
    # Therefore we already apply a mask here, otherwise a possible resampling
    # step by the enduser would lead to errors.
    mapping = mapping.maskedByElevation(1)
    
    return mapping
    
def getMappings(photoTime, cdfL1CacheFolder, cdfL2CacheFolder, 
                altitude=110, maxTimeOffset=2, minBrightness=None, maxBrightness=None,
                offline=False):
    """
    Returns a mapping collection for all available ASIs.
    """
    mappings = []
    
    for station in stations:
        mapping = mappingSingleASI(station, photoTime, cdfL1CacheFolder, cdfL2CacheFolder,
                                   maxTimeOffset=maxTimeOffset, altitude=altitude,
                                   minBrightness=minBrightness, maxBrightness=maxBrightness,
                                   offline=offline)
        if mapping is None:
            continue
        
        mappings.append(mapping)
        
    identifier = 'THEMIS.' + photoTime.strftime('%Y.%m.%d.%H.%M.%S')

    mappingColl = MappingCollection(mappings, identifier, mayOverlap=True)
    return mappingColl


if __name__ == '__main__':
    import auromat.draw; 'OPTIONAL'
    
    cdfL1CacheFolder = '/home/mriecher/data/themis/l1'
    cdfL2CacheFolder = '/home/mriecher/data/themis/l2'
    
    date = datetime(2012,2,4,7,56,26)
    
    mappings = getMappings(date, cdfL1CacheFolder, cdfL2CacheFolder)
    mappings = mappings.maskedByElevation(10)
        
    auromat.draw.saveFig('themis_stereo.png', auromat.draw.drawStereographic(mappings))



# Copyright European Space Agency, 2013

from __future__ import division, print_function, absolute_import

from six.moves import map, _thread
import warnings
import time
import os
import errno
import shutil
import tempfile
import psutil
import futures
import subprocess
import multiprocessing
from functools import partial

from astropy import units as u

import auromat.fits
from auromat.solving.masking import maskStarfield
from auromat.util.image import readFocalLength35mm, loadImage, saveImage
from auromat.util.os import makedirs
import sys

from distutils.version import LooseVersion
from threading import Thread
try:
    import pyfits; 'OPTIONAL'
    PYFITS_TOO_OLD = LooseVersion(pyfits.__version__) < '3.1.0'
except:
    # if unavailable
    PYFITS_TOO_OLD = True

"""
A general purpose astrometry solving module. It has functionality which
allows to solve images that not only contain stars but also other elements.
As it was developed for solving images taken from the ISS towards the earth
(meaning that images contain parts of the earth and may contain spacecraft structures)
it may be biased towards it, yet it very likely is suitable for other kinds
of images as well. 
"""

def solveImages(imagePaths, channel=None, maskingFn=maskStarfield, solveTimeout=60*5, parallel=True, 
                debugOutputFolder=None, noAstrometryPlots=False, pixelError=10, oddsToSolve=None, 
                sigma=None, maxWorkers=None,
                astrometryBinPath=None, useModifiedPath=False, verbose=False):
    """
    Solves multiple images in parallel. See :func:`solveImage` for parameter documentation.
    
    Note: This only works because the actual solving is done in an external
          process (astrometry.net), otherwise nothing would have been gained
          due to Pythons GIL.
    """
    if maxWorkers is None:
        maxWorkers = multiprocessing.cpu_count()
    if parallel and len(imagePaths) > 1:
        workerCount = min(multiprocessing.cpu_count(), maxWorkers)
        
        with futures.ThreadPoolExecutor(max_workers=workerCount) as executor:
            for wcsHeader in executor.map(lambda imagePath: solveImage(imagePath, channel=channel,
                                                                       maskingFn=maskingFn, 
                                                                       solveTimeout=solveTimeout, 
                                                                       debugOutputFolder=debugOutputFolder,
                                                                       noAstrometryPlots=noAstrometryPlots,
                                                                       astrometryBinPath=astrometryBinPath,
                                                                       useModifiedPath=useModifiedPath,
                                                                       pixelError=pixelError,
                                                                       oddsToSolve=oddsToSolve,
                                                                       sigma=sigma,
                                                                       verbose=verbose), 
                                          imagePaths):
                yield wcsHeader
    else:
        for wcsHeader in map(lambda imagePath: solveImage(imagePath, channel=channel,
                                                          maskingFn=maskingFn, 
                                                          solveTimeout=solveTimeout, 
                                                          debugOutputFolder=debugOutputFolder,
                                                          noAstrometryPlots=noAstrometryPlots,
                                                          astrometryBinPath=astrometryBinPath,
                                                          useModifiedPath=useModifiedPath,
                                                          pixelError=pixelError,
                                                          oddsToSolve=oddsToSolve,
                                                          sigma=sigma,
                                                          verbose=verbose), 
                             imagePaths):
            yield wcsHeader

def solveImage(imagePath, channel=None, maskingFn=maskStarfield, sigma=None,
               solveTimeout=60*5, debugOutputFolder=None, noAstrometryPlots=False,
               arcsecRange=None, astrometryBinPath=None, useModifiedPath=False,
               parameters=['xy', 'xy2', 'xy4', 's'], pixelError=10, oddsToSolve=None,
               verbose=False):
    """
    Tries different combinations to solve the image using astrometry.net.
    
    :param imagePath:
    :param channel: the channel to use for source extraction
        'R','G','B', or None for combining all channels into a grayscale image
    :param maskingFn: function to use for masking the given image,
                      if None, masking is skipped
    :param sigma: noise level of the image (optional)
    :param solveTimeout: maximum time in seconds after which astrometry.net is killed
    :param debugOutputFolder: if given, the path to which debug files are written
    :param noAstrometryPlots: whether to let astrometry.net generate plots,
                              if True, then debugOutputFolder must be given 
    :param arcsecRange: tuple(low,high), if not given, then it is guessed from
                        the image file if possible
    :param astrometryBinPath: path to the bin/ folder of astrometry.net;
                              if not given, then whatever is in PATH will be used
    :param bool useModifiedPath: invokes astrometry.net with /usr/bin/env PATH=os.environ['PATH']
                                 This may be useful when the PATH was modified after launching Python.
    :param int pixelError: size of pixel positional error, use higher values (e.g. 10)
                           if image contains star trails (ISS images)
    :param oddsToSolve: default 1e9, see astrometry.net docs
    :rtype: dictionary containing FITS WCS header, or None if solving failed
    """
    imageBase = os.path.splitext(os.path.basename(imagePath))[0]

    tmpDir = tempfile.mkdtemp()
    tmpDirAstrometry = os.path.join(tmpDir, 'astrometry')
    
    # TODO we need to return which solving strategy (downsampling, extractor) was used
    #      -> this should probably be added to the .axy header
    #      as astrometry.net repeatedly downsamples in certain cases it
    #      is hard to determine the correct downsampling that was used
    #      -> see also http://trac.astrometry.net/ticket/1117
        
    if maskingFn is None:
        imageMaskedPath = imagePath
        maskedSuffix = ''
        imageSize = None
        if channel is not None:
            raise NotImplementedError
    else:
        maskedSuffix = '_masked'
        imageMaskedPath = os.path.join(tmpDir, imageBase + maskedSuffix + ".png")
        
        # step 1: create starfield-masked image
        t0 = time.time()
        
        if not debugOutputFolder:
            debugPathPrefix = None
        else:
            debugPathPrefix=os.path.join(debugOutputFolder, imageBase) + '_'
        mask, sigma_ = maskingFn(imagePath, 
                                 debugPathPrefix=debugPathPrefix)
        if sigma is None:
            sigma = sigma_
        im = loadImage(imagePath)
        im[~mask] = 0
        if channel is None:
            pass
        elif channel.lower() == 'r':
            im = im[:,:,0]
        elif channel.lower() == 'g':
            im = im[:,:,1]
        elif channel.lower() == 'b':
            im = im[:,:,2]
        else:
            raise ValueError('channel is "{}" but must be R,G,B or None'.format(channel))
        saveImage(imageMaskedPath, im)
        imageSize = (im.shape[0], im.shape[1])
        
        print('masking:', time.time()-t0, 's')       
        
        if debugOutputFolder is not None:
            shutil.copy(imageMaskedPath, debugOutputFolder)
               
    # step 2: invoke astrometry.net
        
    # Note that astrometry.net assumes that images are taken from earth.
    # As the ISS is very near to earth, the error is most likely below the
    # accuracy that astrometry.net provides and also below the common pixel
    # resolution of images and can therefore be ignored.
    keepTempFiles = False if debugOutputFolder is None else True
    
    if debugOutputFolder is None:
        noAstrometryPlots = True
        
    logFilename = imageBase + maskedSuffix + '.log'
    logFilenameOutput = imageBase + '.log'
    logPath = os.path.join(tmpDirAstrometry, logFilename)
    logBaseOutput = imageBase + '_'
    
    objsBase = imageBase + maskedSuffix + '_objs'
    objsBaseOutput = imageBase + '_objs'
    objsPath = os.path.join(tmpDirAstrometry, objsBase + '.png')
    
    indxBase = imageBase + maskedSuffix + '_indx'
    indxBaseOutput = imageBase + '_indx'
    indxFilename = indxBase + '.png'
    indxFilenameOutput = indxBaseOutput + '.png'
    indxPath = os.path.join(tmpDirAstrometry, indxFilename)
    
    matchFilename = imageBase + maskedSuffix + '.match'
    matchFilenameOutput = imageBase + '.match'
    matchPath = os.path.join(tmpDirAstrometry, matchFilename)
    
    indxXyFilename = imageBase + maskedSuffix + '.xyls'
    indxXyFilenameOutput = imageBase + '.xyls'
    indxXyPath = os.path.join(tmpDirAstrometry, indxXyFilename)
    
    axyFilename = imageBase + maskedSuffix + '.axy'
    axyFilenameOutput = imageBase + '.axy'
    axyPath = os.path.join(tmpDirAstrometry, axyFilename)
    axyBaseOutput = imageBase + '_'

    corrFilename = imageBase + maskedSuffix + '.corr'
    corrFilenameOutput = imageBase + '.corr'
    corrPath = os.path.join(tmpDirAstrometry, corrFilename)
    
    # remove old debug files first
    if keepTempFiles:
        for filename in [matchFilenameOutput, indxFilenameOutput,
                         indxXyFilenameOutput, corrFilenameOutput, axyFilenameOutput, logFilenameOutput,
                         objsBaseOutput + '_xy2.jpg', objsBaseOutput + '_xy4.jpg', objsBaseOutput + '_xy.jpg',
                         objsBaseOutput + '_s.jpg',
                         axyBaseOutput + 'xy2.axy', axyBaseOutput + 'xy4.axy', axyBaseOutput + 'xy.axy',
                         axyBaseOutput + 's.axy',
                         logBaseOutput + 'xy2.log', logBaseOutput + 'xy4.log', logBaseOutput + 'xy.log',
                         logBaseOutput + 's.log']:
            path = os.path.join(debugOutputFolder, filename)
            if os.path.exists(path):
                os.remove(path)
    
    t0 = time.time()
    
    def _copyTempFiles(fitsWcsHeader, name):
        if not keepTempFiles:
            return
        if os.path.exists(objsPath):
            p = os.path.join(debugOutputFolder, objsBaseOutput + '_' + name + '.jpg')
            saveImage(p, loadImage(objsPath)) # convert png to jpg
            
        if not fitsWcsHeader and os.path.exists(axyPath):
            p = os.path.join(debugOutputFolder, axyBaseOutput + name + '.axy')
            shutil.copy(axyPath, p)
            
        if not fitsWcsHeader and os.path.exists(logPath):
            p = os.path.join(debugOutputFolder, logBaseOutput + name + '.log')
            shutil.copy(logPath, p)
    
    # first try astrometry.net's own star extraction (simplexy) with different downsampling options
    
    if arcsecRange:
        if len(arcsecRange) != 2:
            raise ValueError('arcsecRange must be a pair (low,high)')
        arcsecLowHigh = arcsecRange
    else:
        arcsecLowHigh = estimateArcSecRange(imagePath, imageSize)
    _solve = partial(_solveStarfield, imageMaskedPath, tmpDirAstrometry,
                     keepTempFiles=keepTempFiles,
                     timeout=solveTimeout,
                     sigma=sigma,
                     plotsBgImagePath=imagePath,
                     noPlots=noAstrometryPlots,
                     arcsecPerPxLowHigh=arcsecLowHigh,
                     pixelError=pixelError,
                     oddsToSolve=oddsToSolve,
                     astrometryBinPath=astrometryBinPath,
                     useModifiedPath=useModifiedPath,
                     verbose=verbose)
    
    # "The downsampling just seems to generally do a better job of source detection on
    #  most of the images we get, at least at about 2-4. I think it has to do with either
    #  the PSF size (default is too narrow) or saturation (downsampling smears out 
    #  saturated pixels, making them not so saturated)." (Dustin Lang)
    # (https://groups.google.com/d/msg/astrometry/Pp_MZD6s4w8/muuH-1T_zpAJ)
    fitsWcsHeader = None
    if 'xy2' in parameters:
        fitsWcsHeader = _solve(useSextractor=False, downsample=2)    
        _copyTempFiles(fitsWcsHeader, 'xy2')
    
    # Astrometry.net might have already tried downsample=4 when downsample=2 was requested
    # in case no stars could be extracted, so it might happen that we repeat this below.
    # We have to try it anyway because with downsample=2 it could have happened that
    # only 1 or 2 stars in the corners were detected after which astrometry didn't repeat 
    # downsampling and just failed.
    # There should be an astrometry flag to disable the repeated downsampling,
    # see also https://groups.google.com/d/msg/astrometry/qNOgWTL1pVA/BNIbqkXEM-gJ
    # Note that with repeated downsampling, astrometry re-uses the sigma (noise level)
    # from the original downsampling, so there may be slight differences if
    # astrometry is manually called with downsample=4 because it recomputes sigma
    # based on downsample=4. 
    # NOTE: this is the behaviour of the -D flag of image2xy, which cannot be set from solve-field! 
    
    if fitsWcsHeader is None and 'xy4' in parameters:
        if keepTempFiles and os.path.exists(tmpDirAstrometry):
            shutil.rmtree(tmpDirAstrometry)
        fitsWcsHeader = _solve(useSextractor=False, downsample=4)          
        _copyTempFiles(fitsWcsHeader, 'xy4')
                    
    # if this didn't work, we try SExtractor for star extraction
    
    if fitsWcsHeader is None and 's' in parameters:
        if keepTempFiles and os.path.exists(tmpDirAstrometry):
            shutil.rmtree(tmpDirAstrometry)
        fitsWcsHeader = _solve(useSextractor=True, downsample=None)  
        _copyTempFiles(fitsWcsHeader, 's')
            
    # in case the input image has a low resolution (which is not the case for ISS images)
    # we also might have luck with downsampling disabled and using simplexy:
    
    if fitsWcsHeader is None and 'xy' in parameters:
        if keepTempFiles and os.path.exists(tmpDirAstrometry):
            shutil.rmtree(tmpDirAstrometry)
        fitsWcsHeader = _solve(useSextractor=False, downsample=None)        
        _copyTempFiles(fitsWcsHeader, 'xy')
        
    print('solving:', time.time()-t0, 's')
    
    if fitsWcsHeader is not None:
        if keepTempFiles and os.path.exists(indxPath):
            p = os.path.join(debugOutputFolder, indxBaseOutput + '.jpg')
            saveImage(p, loadImage(indxPath)) # convert png to jpg
        
        tempPaths = [matchPath, indxXyPath, axyPath, corrPath, logPath]
        outputFilenames = [matchFilenameOutput, indxXyFilenameOutput, axyFilenameOutput, 
                           corrFilenameOutput, logFilenameOutput]
        if keepTempFiles:
            outputPaths = [os.path.join(debugOutputFolder,f) for f in outputFilenames]
            for tempPath, outputPath in zip(tempPaths, outputPaths):    
                if os.path.exists(tempPath):
                    shutil.move(tempPath, outputPath)
                else:
                    print('WARNING!! {} does not exist after successfully solving, but it should!'.format(tempPath), file=sys.stderr)
                        
    shutil.rmtree(tmpDir)
        
    return fitsWcsHeader

def estimateArcSecRange(imagePath, imageSize=None):
    focal35mm = readFocalLength35mm(imagePath)
    if focal35mm is None:
        return None
    focal35mmLow = focal35mm*0.9
    focal35mmHigh = focal35mm*1.1
    if imageSize is None:
        _,w, = loadImage(imagePath).shape
    else:
        _,w = imageSize
    # see http://cbellh47.blogspot.nl/2010/01/astrometry-101-pixel-scale.html
    pixelSizeMm = 35/w
    arcsecLow = (pixelSizeMm/focal35mmHigh * u.rad).to(u.arcsec).value
    arcsecHigh = (pixelSizeMm/focal35mmLow * u.rad).to(u.arcsec).value
    return (arcsecLow, arcsecHigh)

def _solveStarfield(imagePath, tmpDir=None, keepTempFiles=False, timeout=60*1, 
                    useSextractor=True, downsample=2, sigma=None, 
                    searchField=None, arcsecPerPxLowHigh=None, 
                    pixelError=10, oddsToSolve=None,
                    plotsBgImagePath=None, noPlots=False,
                    astrometryBinPath=None, useModifiedPath=False, verbose=False):
    """
    NOTE: The astrometry/bin folder must be in the PATH. 
    
    :param imagePath:
    :param timeout: time in seconds after which the solving process is terminated
                    If solving fails, increase this. astrometry.net tries to solve
                    using bright stars first, and if these are not in the center then
                    the distortion may be too high and only fainter stars might lead
                    to a successful solve, which in turn needs longer processing time.
    :param useSextractor: sextractor sometimes delivers better results than
                          the built-in star extraction of astrometry.net (image2xy)
    :param int|None downsample: Whether and how much astrometry should downsample the image before solving
    :param sigma: noise level override
    :param (ra,dec,radius)|None searchField: search only within 'radius' of the field center 'ra','dec', all in degrees
    :param str plotsBgImagePath: path to .jpg file to use as background for all plots
    :param tuple arcsecPerPxLowHigh: tuple of lower and upper arcsecPerPx to restrict search
    :param int pixelError: size of pixel positional error, use higher values (e.g. 10)
                           if image contains star trails (ISS images)
    :param oddsToSolve: default 1e9, see astrometry.net docs
    :param bool useModifiedPath: invokes astrometry.net with /usr/bin/env PATH=os.environ['PATH']
                                 This may be useful when the PATH was modified after launching Python, Unix only.
    :rtype: wcs header, or None if no solution was found
    """
    # adapt these constants if necessary for newer astrometry.net versions
    solvefieldName = 'solve-field'
    backendName = 'astrometry-engine'
    
    if tmpDir is None and keepTempFiles is True:
        print("solveStarfield: tmpDir is not set but keepTempFiles is true, this doesn't make much sense")
    
    if tmpDir is None:
        tmpDir = tempfile.mkdtemp()
    tmpTmpDir = os.path.join(tmpDir, "tmp")
    makedirs(tmpTmpDir)
        
    imageBase = os.path.splitext(os.path.basename(imagePath))[0]
        
    solvedPath = os.path.join(tmpDir, imageBase + ".solved")
    wcsPath = os.path.join(tmpDir, imageBase + ".wcs")
    matchPath = os.path.join(tmpDir, imageBase + ".match")
    indxXyPath = os.path.join(tmpDir, imageBase + ".xyls")
    corrPath = os.path.join(tmpDir, imageBase + ".corr")
    logPath = os.path.join(tmpDir, imageBase + ".log")
    
    if not astrometryBinPath:
        astrometryBinPath = ''
    
    args = [os.path.join(astrometryBinPath, solvefieldName)]
    
    if useModifiedPath:
        args = ['/usr/bin/env', 'PATH=' + os.environ['PATH']] + args
    
    args += ["--cpulimit", str(timeout)] # see https://github.com/dstndstn/astrometry.net/issues/6
    args += ["--dir", tmpDir, "--temp-dir", tmpTmpDir, "--no-delete-temp"]
    # TODO there are no params for the plot filenames (..-indx.png and ..-objs.png)
    #      and also not for .axy
    args += ["--wcs", wcsPath, "--solved", solvedPath, "--match", matchPath]
    args += ["--index-xyls", indxXyPath, "--corr", corrPath]
    args += ["--crpix-center"]
    args += ["--no-background-subtraction"]
    
    if PYFITS_TOO_OLD:
        args += ["--no-remove-lines"]
        args += ["--no-fits2fits"]

    if arcsecPerPxLowHigh is not None:
        arcsecLow, arcsecHigh = arcsecPerPxLowHigh
        args += ["--scale-low", str(arcsecLow), "--scale-high", str(arcsecHigh), "--scale-units", "arcsecperpix"]
    
    args += ["--no-tweak"] # no SIP polynomial; we correct lens distortion before-hand
    args += ["--pixel-error", str(pixelError)]
    
    if oddsToSolve:
        args += ["--odds-to-solve", str(oddsToSolve)]
    
    if verbose:
        args += ["--verbose"]
    
    if sigma:
        args += ["--sigma", str(sigma)]
    
    if searchField:
        ra,dec,radius = searchField
        args += ["--ra", str(ra), "--dec", str(dec), "--radius", str(radius)]
    
    if downsample:
        args += ["--downsample", str(downsample)]
    
    if useSextractor:
        args += ["--use-sextractor"]

    if plotsBgImagePath:
        args += ["--plot-bg", plotsBgImagePath]
        
    if keepTempFiles:
        print('astrometry files in', tmpDir)
        if noPlots:
            args += ["--no-plots"]
    else:
        args += ["--no-plots"]
        args += ["--new-fits", "none", "--index-xyls", "none", "--rdls", "none", "--corr", "none"]
        
    args += [imagePath]

    print(' '.join(args))
    
    def print_and_store_output(out, logPath):
        # work-around the fact that using sys.stdout in subprocess breaks
        # if sys.stdout got redirected to a non-file object (e.g. StringIO)
        with open(logPath, 'w') as logfile:
            logfile.write(' '.join(args) + '\n')
            for line in iter(out.readline, b''):
                print(line, end='')
                logfile.write(line)
            out.close()
    
    try:
        process = psutil.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=1)
        t = Thread(target=print_and_store_output, args=(process.stdout, logPath))
        t.daemon = True # thread dies with the program, should not be needed if no error occurs
        t.start()
        
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise RuntimeError('The "' + solvefieldName + '" program from astrometry.net could not be launched. ' + \
                               'Make sure it is in the PATH!')
        else:
            raise
     
    try:
        process.wait(timeout+10)
    except: # psutil.TimeoutExpired or KeyboardInterrupt (Ctrl+C)
        print('astrometry.net timeout reached, killing processes now')
        
        backend = list(filter(lambda p: p.name() == backendName, process.get_children(True)))
        processes = [process] + backend
        
        for p in processes:
            print('terminating {} (pid {})'.format(p.name(), p.pid))
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                pass
            
        # if astrometry is just writing its solution, let it have some time to do so
        t = 5
        _, alive = psutil.wait_procs(processes, t)
        if alive:
            # This shouldn't happen.
            warnings.warn('solve-field or backend did NOT exit in ' + t + 's after SIGTERM was sent!' +
                          '...killing them now and ignoring results')
            for p in processes:
                print('killing {} (pid {})'.format(p.name(), p.pid))
                try:
                    p.kill()
                except psutil.NoSuchProcess:
                    pass
            if not keepTempFiles:
                shutil.rmtree(tmpDir)
            return None

    # FIXME astrometry seems to leave files in an inconsistent state 
    #       if it is terminated while persisting results
    #       -> e.g. it seems that the .solved file is written first, and then the rest
    #  -> quick work-around: check for .solved AND .wcs, and catch wcs reading exceptions
    #     may still fail in the end, but the probability is lower
    #  in theory, astrometry should handle SIGTERM accordingly instead of just exiting immediately
       
    try:
        if not os.path.exists(solvedPath) or not os.path.exists(wcsPath):
            return None
        fitsWcsHeader = auromat.fits.readHeader(wcsPath)
    except Exception as e:
        warnings.warn('error reading wcs file ' + wcsPath)
        print(repr(e))
        return None
    finally:
        if not keepTempFiles:
            shutil.rmtree(tmpDir)
        
    return fitsWcsHeader

    


# Copyright European Space Agency, 2013

"""
The module contains algorithms to automatically mask those image areas
which are most likely containing the starfield.

It does so by using a combination of constant thresholding, adaptive thresholding,
line detection and contour detection. It assumes that the starfield has a more or
less constant background brightness. It was developed and tested with ISS aurora
images, so it may be biased towards those.
"""

from __future__ import division, print_function, absolute_import

import os
import traceback
from collections import namedtuple

from math import pi
import numpy as np
from scipy.signal import convolve2d
import cv2 as cv

try:
    IMWRITE_PNG_COMPRESSION = cv.IMWRITE_PNG_COMPRESSION
    IMWRITE_JPEG_QUALITY = cv.IMWRITE_JPEG_QUALITY
except AttributeError:
    IMWRITE_PNG_COMPRESSION = cv.cv.CV_IMWRITE_PNG_COMPRESSION
    IMWRITE_JPEG_QUALITY = cv.cv.CV_IMWRITE_JPEG_QUALITY

from auromat.draw import drawHistogram, saveFig

from auromat.solving.noiseestimation import _estimateNoiseLevel as _doEstimateNoiseLevel
from auromat.solving.viewasblocks import view_as_blocks
  

# Paradigm: The less false stars, the higher the solving probability and
#           the higher the accuracy.
# When there are less false stars then it's much easier for astrometry.net to tweak
# a solution (=align the stars with other stars of the image except the already matched
# stars from the quad/triangle). 

def maskStarfieldRect(imagePath, topLeft, bottomRight, debugPathPrefix=None):
    """
    Mask an image using the given coordinates representating a rectangle.
    
    :param imagePath:
    :param topLeft: (x,y) pair of top left rectangle pixel
    :param bottomRight: (x,y) pair of bottom right rectangle pixel
    :rtype: tuple (mask, sigma)
    """
    if isinstance(imagePath, np.ndarray):
        im = imagePath
    else:
        im = cv.imread(imagePath)
    h,w = im.shape[0], im.shape[1]
    
    x1,y1 = topLeft
    x2,y2 = bottomRight
    
    mask = np.zeros((h,w), np.bool)
    mask[y1:y2+1,x1:x2+1] = True
    
    sigma = _estimateNoiseLevel(im[y1:y2+1,x1:x2+1,0])
    
    return mask, sigma

def _binarizeStarfieldImage(imgray, fudge=20):
    """
    
    :param imgray: grayscale image as returned by cv.cvtColor(im, cv.COLOR_BGR2GRAY)
    :rtype: binary image where stars are retained
    """
    maxThreshold = 150
    
    # Find the average background brightness of the starfield.
    # In most cases this corresponds to the first spike of the image histogram,
    # assuming that the starfield background is the darkest part of the image.
    # The threshold for binarization is then the average plus an empirical value.
    hist = cv.calcHist([imgray],[0],None,[256],[0,255]).reshape(256)
    # Sometimes there are small bumps before the first main spike.
    # To prevent choosing the wrong spike, the histogram is slightly smoothed.
    hist[1:-1] = (hist[:-2] + hist[1:-1] + hist[2:]) / 3 # smoothing with window=3
    histDiff = hist[1:] - hist[:-1]
    firstSpike = np.argmax(histDiff<0)
    threshold = min(firstSpike + fudge, maxThreshold)
        
    _,binary = cv.threshold(imgray, threshold, 255, cv.THRESH_BINARY)
    return binary, hist, threshold, firstSpike
    
def _findAndCategorizeContours(binary):
    # see auromat.utils.outline_opencv for why we need to pad the image
    imCv = np.zeros((binary.shape[0]+2, binary.shape[1]+2), dtype=np.uint8)
    imCv[1:-1,1:-1] = binary
    contours,_ = cv.findContours(imCv, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)    
    contours = np.asarray(contours)
    contours -= 1

    area = np.asarray([cv.contourArea(c) for c in contours])
    rectAxes = np.asarray([cv.minAreaRect(c)[1] for c in contours])

    # isBigContour needs to be big enough to not discard bigger stars!
    # this could leave some spacecraft structures intact which may or may not confuse astrometry    
    # TODO the ratio below should depend on the estimated celestial pixel scale
    #      and the exposure time (longer exposure = longer star trails) 
    bigContourAreaRatio = 0.000013 # 0.0013% of the image area (~160 pixels for 12MP images)
    bigContourArea = bigContourAreaRatio*(binary.shape[0]*binary.shape[1])
    isBigContour = area > int(bigContourArea)
    isSmallContour = ~isBigContour
    
    longRatioThreshold = 5
    with np.errstate(divide='ignore', invalid='ignore'): # division produces nans and infs
        rectRatio = rectAxes[:,0]/rectAxes[:,1] if len(contours) > 0 else np.array([])

    with np.errstate(invalid='ignore'):
        isLongContour = np.logical_and(area > 20, # exclude very tiny long contours (could be stars) and inf ratios
                                       np.logical_or(rectRatio > longRatioThreshold, 
                                                     rectRatio < 1/longRatioThreshold)
                                       )
    isSmallLongContour = np.logical_and(isSmallContour, isLongContour)
    isSmallShortContour = np.logical_and(isSmallContour, ~isLongContour)
    
    return contours, area, isBigContour, isSmallLongContour, isSmallShortContour

def _getBlockShape(im):
    # assumes landscape images
    
    # roughly square blocks
    blocksX = 16
    blocksY = 12
    
    if im.shape[0] % blocksY != 0:
        # if 12 doesn't work, we use 8, assuming that height is divisible by 16
        blocksY = 8
        
    if im.shape[0] % blocksY != 0 or im.shape[1] % blocksX != 0:
        raise NotImplementedError('(width, height) of image must be divisible by (' +
                                  str(blocksX) + ',' + str(blocksY) + ') for block masking, ' +
                                  str(im.shape[1]) + 'x' + str(im.shape[0]))
    
    blockH = im.shape[0]//blocksY
    blockW = im.shape[1]//blocksX    
    return (blockH, blockW)

def _createStarfieldMask(im, contours, area, isBigContour, isSmallLongContour, 
                         blackenLowerPart = True):
    mask = np.ones((im.shape[0], im.shape[1]), np.bool)
    
    blockH, blockW = _getBlockShape(im)
    
    if blackenLowerPart:
        # 1. find biggest contour
        # 2. check if the top end is within the bottom two thirds and the lower end
        #    within the bottom half of the image
        # 3. if 2 is satisfied, paint everything black from the top of the contour to the bottom of the image
        #    Note: the assumption is that such a big chunk is likely be a part of the earth
        # 4. if 2 is not satisfied, paint the lower half of the image black (fallback)
        # This procedure helps in case the image is very dark and faint citylights would be recognized as stars.
        biggestContour = contours[np.argmax(area)]
        _,y,_,h = cv.boundingRect(biggestContour)
        if y > im.shape[0]/3 and y+h > im.shape[0]/2:
            fromy = y
        else:
            fromy = im.shape[0]//2
        
        # cut off below the corresponding block
        fromyBlockBoundary = int(np.ceil(fromy/blockH)*blockH)
        mask[fromyBlockBoundary:] = False


    # devide image into blocks and paint those blocks black which contain big or (small and long)
    # contours, as these are very likely spacecraft structures
    assert isBigContour is not None or isSmallLongContour is not None
    if isSmallLongContour is None:
        isOffendingContour = isBigContour
    else:
        isOffendingContour = np.logical_or(isBigContour, isSmallLongContour)
    
    # draw and fill all offending contours on a blank binary image
    # then for each block check if there are pixels which are black
    imFilledOffenders = np.zeros(mask.shape, np.uint8)
    cv.fillPoly(imFilledOffenders, contours[isOffendingContour], 255)

    blockViewMask = view_as_blocks(mask, (blockH,blockW))
    blockViewOffenders = view_as_blocks(imFilledOffenders, (blockH,blockW))

    isBlockContainingOffenders = (blockViewOffenders==255).any(axis=-1).any(axis=-1)
    blockViewMask[isBlockContainingOffenders] = False

    return mask

def _masked_adaptive_threshold(image,mask,max_value,size,C):
    """
    thresholds only using the unmasked pixels
    Note: becomes slightly inaccurate on very dark areas
    Note: image needs to be black at masked pixels
    
    adapted from http://stackoverflow.com/a/10551103
    """
    mask = mask.astype(np.uint8) * 255
    conv = cv.blur(image, (size, size)).astype(float)
    number_neighbours = cv.blur(mask, (size, size)).astype(float)
    with np.errstate(invalid='ignore'):
        image = image-255*(conv/number_neighbours)    
    binary = np.zeros_like(image, dtype=np.uint8)
    binary[np.logical_and(image > -C, mask)] = max_value
    return binary

def maskStarfield(imagePath, channel=None, blackenLowerPart=True, ignoreVeryDark=True,
                  debugPathPrefix=None, debugJpegQuality=80):
    """
    Automatic masking of the starfield in the given image using
    a combination of image processing and object detection steps.
    
    :param str channel: the channel to use for analysis
        'R','G','B', or None for combining all channels into a grayscale image
    :param bool blackenLowerPart: If the earth is in the lower part of the image
                             then this should be set to True as it will
                             broadly mask parts of the earth not detected otherwise.
    :param bool ignoreVeryDark: If True, then areas (block) which are almost totally black
                           are not considered as starfield. This is sometimes useful
                           to ignore very dark spacecraft structures which would later
                           on be detected as stars.
    :param str debugPathPrefix: if given, the folder in which to store debug images
                                illustrating the different processing stages
    :param int debugJpegQuality: JPEG quality from 0 to 100 used for storing debug images;
                                 only contour images are currently saved as JPEG
    :rtype: tuple (mask, sigma)
    """    
    if debugPathPrefix:
        red = (0,0,255)
        green = (0,255,0)
        orange = (0,106,255)

        debugHistogramPath = debugPathPrefix + 'hist.svg'
        debugThresholdedImagePath = debugPathPrefix + 'thresh.png'
        debugContoursImagePath = debugPathPrefix + 'cont.jpg'
        debugContoursMaskImagePath = debugPathPrefix + 'cont_mask.jpg'
        debugAdaptiveThresholdedImagePath = debugPathPrefix + 'thresh_adapt.png'
        debugCutoffImagePath = debugPathPrefix + 'cutoff.jpg'
    
    if isinstance(imagePath, np.ndarray):
        # assume RGB image array
        im = np.require(imagePath, np.uint8, 'C')
        im = cv.cvtColor(im, cv.COLOR_RGB2BGR)
        imagePath = '[array]'
    else:
        im = cv.imread(imagePath)
    if channel is None:
        imgray = cv.cvtColor(im, cv.COLOR_BGR2GRAY)
    elif channel.lower() == 'r':
        imgray = im[:,:,2]
    elif channel.lower() == 'g':
        imgray = im[:,:,1]
    elif channel.lower() == 'b':
        imgray = im[:,:,0]
    else:
        raise ValueError('channel is "{}" but must be R,G,B or None'.format(channel))
    
    # opencv requires a contiguous array..
    imgray = np.require(imgray, np.uint8, 'C')
    
    # Step 1: Find dark areas which might be starfield
    
    fudge = 20
    binary, hist, threshold, firstSpike = _binarizeStarfieldImage(imgray, fudge=fudge)
    contours, area, isBigContour, isSmallLongContour, isSmallShortContour = _findAndCategorizeContours(binary)
    mask = _createStarfieldMask(im, contours, area, isBigContour, None, blackenLowerPart)
    starfieldAreaRatio = np.sum(mask) / (mask.shape[0]*mask.shape[1])
        
    while starfieldAreaRatio < 0.1:
        # A part other then the starfield was probably picked as reference threshold because it was darker.
        # Remember that the first spike in the histogram (=darkest part) is used to determine the threshold.
        # To fix this situation, we just raise the threshold and hope for the best.
        print('Starfield area is only ' + "{0:.2f}".format(starfieldAreaRatio*100) +\
              '% (< 10%). Trying a higher threshold. ' +\
              '(' + os.path.basename(imagePath) + ')')
        fudge += 20
        binary, hist, threshold, firstSpike = _binarizeStarfieldImage(imgray, fudge=fudge)
        contours, area, isBigContour, isSmallLongContour, isSmallShortContour = _findAndCategorizeContours(binary)
        # FIXME use small long contours as well for masking, why did we disable that?
        mask = _createStarfieldMask(im, contours, area, isBigContour, None, blackenLowerPart)
        
        starfieldAreaRatio = np.sum(mask) / (mask.shape[0]*mask.shape[1])
        if starfieldAreaRatio >= 0.1:
            print('Starfield area is now ' + "{0:.2f}".format(starfieldAreaRatio*100) + '%')
        elif fudge > 100:
            print('giving up')
            break
        
    if debugPathPrefix and debugHistogramPath:
        vlines = [(firstSpike, 'red'), (threshold, 'blue')]
        try:
            saveFig(debugHistogramPath, drawHistogram(hist, vlines, 
                                                      xlabel='Intensity', 
                                                      ylabel='Pixel Count',
                                                      linecolor='black'))
        except:
            ex = traceback.format_exc()
            with open(debugPathPrefix + 'matplotlib.EXCEPTION', 'w') as fp:
                fp.write(ex)
        
    if debugPathPrefix and debugThresholdedImagePath:
        cv.imwrite(debugThresholdedImagePath, binary, [IMWRITE_PNG_COMPRESSION, 9])
        
    if debugPathPrefix and debugContoursImagePath:
        imContours = im.copy()
        cv.drawContours(imContours,contours[isBigContour],-1,red,2)
        cv.drawContours(imContours,contours[isSmallLongContour],-1,orange,2)
        cv.drawContours(imContours,contours[isSmallShortContour],-1,green,2)
        cv.imwrite(debugContoursImagePath, imContours, [IMWRITE_JPEG_QUALITY, debugJpegQuality])
        del imContours

    if debugPathPrefix and debugContoursMaskImagePath:
        # draw the borders of the current mask onto the original image
        imContoursMask = im.copy()
        res = _findAndCategorizeContours(mask)
        contours = res[0]
        cv.drawContours(imContoursMask,contours,-1,(255,255,255),3)
        cv.imwrite(debugContoursMaskImagePath, imContoursMask, [IMWRITE_JPEG_QUALITY, debugJpegQuality])
        del imContoursMask

    imgray[~mask] = 0
    
    # Step 2: Filter out dark areas which are probably not starfield
    
    # Step 2a: try to find lines and mask blocks containing them
    binary = _masked_adaptive_threshold(imgray, mask, 255, 89, -1)
    if debugPathPrefix and debugAdaptiveThresholdedImagePath:
        binaryC = cv.cvtColor(binary, cv.COLOR_GRAY2BGR)

    binary = cv.medianBlur(binary, 3)
    lines = cv.HoughLinesP(binary.copy(), 1, pi/180, 200, minLineLength=100, maxLineGap=4)
    
    blockShape = _getBlockShape(im)
    blockH, blockW = blockShape
    blockViewMask = view_as_blocks(mask, blockShape)
    
    if lines is not None:        
        # draw all lines on a blank binary image
        # then for each block check if it contains part of a line
        imFilledOffenders = np.zeros(mask.shape, np.uint8)
        
        for line in lines[0,:]:
            cv.line(imFilledOffenders, (line[0], line[1]), (line[2], line[3]), 255)
            
            if debugPathPrefix and debugAdaptiveThresholdedImagePath:
                cv.line(binaryC, (line[0], line[1]), (line[2], line[3]), red, 5)

        blockViewOffenders = view_as_blocks(imFilledOffenders, blockShape)
    
        isBlockContainingOffenders = (blockViewOffenders==255).any(axis=-1).any(axis=-1)
        blockViewMask[isBlockContainingOffenders] = False
        
    if debugPathPrefix and debugAdaptiveThresholdedImagePath:
        cv.imwrite(debugAdaptiveThresholdedImagePath, binaryC, [IMWRITE_PNG_COMPRESSION, 9])
       
    # Step 2b: make very dark pixels black and mask all-black blocks
    if ignoreVeryDark:
        if debugPathPrefix and debugCutoffImagePath:
            wasStarfieldBlock = blockViewMask.all(axis=-1).all(axis=-1)
        
        imgrayCutoff = imgray.copy()
        # blurring will wash out very tiny artifacts (which could also be faint small stars)
        # this helps here because we require that absolutely all of a block must be black
        imgrayCutoff = cv.blur(imgrayCutoff, (3,3))
        cutoffThreshold = max(30, firstSpike + 20)
        print('cutoffThreshold:', cutoffThreshold, os.path.basename(imagePath))
        imgrayCutoff[imgrayCutoff<cutoffThreshold] = 0
    
        blockViewCutoff = view_as_blocks(imgrayCutoff, blockShape)
        isBlockPureBlack = (blockViewCutoff==0).all(axis=-1).all(axis=-1)
        
        if debugPathPrefix and debugCutoffImagePath:
            imCutoff = im.copy()
            imCutoff[~mask] = 0
            
            blockViewCutoffC = view_as_blocks(imCutoff, (blockH,blockW,3))
            blockGotMasked = np.logical_and(wasStarfieldBlock, isBlockPureBlack)
            # draw rectangle around each block that was masked
            blockViewCutoffC[blockGotMasked,0,:4,:] = red
            blockViewCutoffC[blockGotMasked,0,-4:,:] = red
            blockViewCutoffC[blockGotMasked,0,:,:4] = red
            blockViewCutoffC[blockGotMasked,0,:,-4:] = red
            cv.imwrite(debugCutoffImagePath, imCutoff, [IMWRITE_JPEG_QUALITY, debugJpegQuality])
            del imCutoff
        
        blockViewMask[isBlockPureBlack] = False
        
    # Step 3: Filter out lonely starfield blocks (surrounded by non-starfield blocks)
    isStarfieldBlock = blockViewMask.all(axis=-1).all(axis=-1)
    
    neighborCountKernel = np.ones((3,3), dtype=int)
    neighborCountKernel[1,1] = 0
    neighbors = convolve2d(isStarfieldBlock.astype(int), neighborCountKernel, mode='same')
    isLonelyBlock = np.logical_and(isStarfieldBlock, neighbors == 0)
    blockViewMask[isLonelyBlock] = False
       
    # estimate noise level using the biggest starfield rectangle we got
    (rectY,rectX), (rectH,rectW) = _max_size_rectangle(isStarfieldBlock, value=True)
    rectY, rectH = rectY*blockH, rectH*blockH
    rectX, rectW = rectX*blockW, rectW*blockW
    imgrayBiggestRect = imgray[rectY:rectY+rectH, rectX:rectX+rectW]
    sigma = _estimateNoiseLevel(imgrayBiggestRect)
    print('Sigma:', sigma)
    if debugPathPrefix:
        with open(os.path.join(debugPathPrefix + '.sigma'), 'w') as fp:
            fp.write(str(sigma))

    return mask, sigma

def _estimateNoiseLevel(imgray):
    sigma = _doEstimateNoiseLevel(imgray)
    # sigma seems to be often lower than what astrometry.net calculates
    # NOTE: this is a dirty HACK, sometimes sigma is too high
    sigma = max(0.9, sigma * 2.5)
    return sigma

def _max_size_rectangle(mat, value=True):
    """Return (row,column),(height, width) of the largest rectangle containing 
    all `value`'s.
    """
    # adapted from https://gist.github.com/zed/776423
    
    #The original version was modified such that also the position of the largest
    #rectangle is returned, in addition to its size.
    #
    #Copyright (c) 2014, zed <isidore.john.r@gmail.com>
    #
    #Permission to use, copy, modify, and/or distribute this software for any
    #purpose with or without fee is hereby granted, provided that the above
    #copyright notice and this permission notice appear in all copies.
    #
    #THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
    #WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
    #MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
    #ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
    #WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
    #ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
    #OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
    area = np.product
    it = iter(mat)
    hist = [(el==value) for el in next(it, [])]
    max_size, max_column_index = _max_rectangle_size_hist(hist)
    max_row_index_bottom = 0
    for bottom_row_index, row in enumerate(it):
        hist = [(1+h) if el == value else 0 for h, el in zip(hist, row)]
        row_max_size, row_max_column_index = _max_rectangle_size_hist(hist)
        if area(row_max_size) > area(max_size):
            max_size = row_max_size
            max_row_index_bottom = bottom_row_index + 1
            max_column_index = row_max_column_index
    return (max_row_index_bottom-max_size[0]+1, max_column_index), max_size 

def _max_rectangle_size_hist(histogram):
    """Return size and column of the largest rectangle that fits entirely under
    the histogram.
    """
    # see _max_size_rectangle for origin and license text
    
    # The original version was modified such that also the column of the largest
    # rectangle is returned, in addition to its size.    
    Info = namedtuple('Info', 'start height')
    area = np.product
    stack = []
    top = lambda: stack[-1]
    max_size = (0, 0) # height, width of the largest rectangle
    max_column_index = 0
    pos = 0 # current position in the histogram
    for pos, height in enumerate(histogram):
        start = pos # position where rectangle starts
        while True:
            if not stack or height > top().height:
                stack.append(Info(start, height)) # push
            elif stack and height < top().height:
                current_size = (top().height, (pos - top().start))
                if area(current_size) > area(max_size):
                    max_size = current_size
                    max_column_index = top().start
                start, _ = stack.pop()
                continue
            break # height == top().height goes here
    pos += 1    
    for start, height in stack:
        if area((height, (pos - start))) > area(max_size):
            max_size = (height, (pos - start))
            max_column_index = start
    return max_size, max_column_index



"""
This package contains modules to determine coordinates of space objects 
(:mod:`~auromat.coordinates.spacetrack` and :mod:`~auromat.coordinates.ephem`), 
to convert existing coordinates into other reference frames 
(:mod:`~auromat.coordinates.transform` and :mod:`~auromat.coordinates.igrf`),
to calculate intersection points between a ray and an ellipsoid
(:mod:`~auromat.coordinates.intersection`), and to perform geodesic
calculations (:mod:`~auromat.coordinates.geodesic`).

This package does not depend on the `mapping` objects used in other parts of
the :mod:`auromat` package and can therefore be re-used generically for
other purposes.
"""


#Copyright (c) 2012, Tolga Birdal
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are
#met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the distribution
#    * Neither the name of the Gravi Information Technologies and Consultancy Ltd nor the names
#      of its contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
#LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#POSSIBILITY OF SUCH DAMAGE.

from math import sqrt, pi
import numpy as np
from scipy.ndimage.filters import correlate1d


def _estimateNoiseLevel(imgray):
    """
    Estimates the noise level of the given one-channel image.
    
    This code is adapted from
    http://www.mathworks.com/matlabcentral/fileexchange/36941-fast-noise-estimation-in-images
    
    The original description follows:
    
        by Tolga Birdal
        
        This is an extremely simple m-file which implements the method described
        in : J. Immerkaer, "Fast Noise Variance Estimation", Computer Vision and
        Image Understanding, Vol. 64, No. 2, pp. 300-302, Sep. 1996
                
        The advantage of this method is that it includes a Laplacian operation 
        which is almost insensitive to image structure but only depends on the 
        noise in the image. 
    """    
    h,w = imgray.shape[0], imgray.shape[1]
    
    # compute sum of absolute values of Laplacian
    kernel = np.array([1,-2,1])
    conv = correlate1d(imgray.astype(np.float), kernel, axis=0)
    conv = correlate1d(conv, kernel, axis=1)
    sigma = np.sum(np.abs(conv))
    
    # scale sigma with proposed coefficients
    sigma = sigma*sqrt(0.5*pi)/(6*(w-2)*(h-2))
    return sigma


# Copyright European Space Agency, 2013

"""
A module for reading and writing specific FITS header values.
"""

from __future__ import division, absolute_import, print_function

from six import string_types
import warnings
from datetime import datetime, timedelta
import numpy as np
from math import sqrt, sin, cos, atan2
from astropy.io import fits
from astropy.wcs.wcs import WCS
import astropy.units as u
import astropy.coordinates as coord
import time
import sys
import astropy.utils.data
from astroquery.query import suspend_cache

try:
    from astroquery.vizier import Vizier
except ImportError as e:
    print(str(e))
    warnings.warn('astroquery not available, .fits.getCatalogStars cannot be used')

def readHeader(filePath):
    """ Return the primary FITS header. """
    return fits.getheader(filePath, checksum=True)
    
def writeHeader(filePath, header, overwrite=False):
    """
    Creates a new file containing the given FITS header.
    
    :param filePath: Path where FITS header will be written
    :param header: FITS header to write
    :param overwrite: Raises an error if False and filePath already exists
    """
    fits.writeto(filePath, None, header, clobber=overwrite, checksum=True)
    
def getPixelScale(header):
    """
    Returns the pixel scale in degrees/pixel based on the CD matrix of the WCS header.
    SIP distortion coefficients (if present) are not considered.
    """
    assert header['CUNIT1'] == 'deg' and header['CUNIT2'] == 'deg'
    cd11 = header['CD1_1']
    cd21 = header['CD2_1']
    scale = sqrt(cd11**2 + cd21**2)
    return scale

def getRotationAngle(header):
    """
    Returns the rotation (roll) angle in degrees based on the CD matrix of the WCS header.
    Angle is in [-180,180].
    """
    assert header['CUNIT1'] == 'deg' and header['CUNIT2'] == 'deg'
    cd11 = header['CD1_1']
    cd21 = header['CD2_1']

    rho_a = atan2(cd21, cd11)
    
    return np.rad2deg(rho_a)

def cd11cd21(scale, rotation):
    """
    Calculates CD11 and CD21 from the given pixel scale and rotation.
    
    :param scale: pixel scale in degrees/pixel
    :param rotation: rotation angle in degrees within [-180,180]
    :rtype: tuple (cd11, cd21)
    """
    rho = np.deg2rad(rotation)
    cd11 = scale * cos(rho)
    cd21 = scale * sin(rho)
    return cd11, cd21

def setCdMatrix(header, scale, rotation):
    """
    Sets the CD matrix from the given pixel scale and rotation.
    
    :param header:
    :param scale: pixel scale in degrees/pixel
    :param rotation: rotation angle in degrees within [-180,180]
    """
    cd11, cd21 = cd11cd21(scale, rotation)
    header['CD1_1'] = cd11
    header['CD1_2'] = -cd21
    header['CD2_1'] = cd21
    header['CD2_2'] = cd11

def getRadius(header, extend=0):
    """
    Returns the radius in degrees of the circle which encloses the image.
    Degrees are based on the CD matrix of the WCS header. 
    SIP distortion coefficients (if present) are not considered.
    
    :param header: must also contain IMAGEW and IMAGEH
    :param extend: how much to extend the circle in percent [0,1]
    """
    diagPx = sqrt(header['IMAGEW']**2 + header['IMAGEH']**2)
    radiusPx = diagPx/2 * (1+extend)
    radiusDeg = getPixelScale(header) * radiusPx
    return radiusDeg

def getCenterRADec(header):
    """
    Returns RA,Dec for the image center.
    
    :param header: must also contain IMAGEW and IMAGEH
    :rtype: tuple (ra,dec) in degrees, ra is [0,360], dec is [-90,90]
    """
    w = header['IMAGEW']
    h = header['IMAGEH']
    del header['NAXIS'] # emits a warning in WCS constructor if present (as it's 0 for .wcs files)
    return WCS(header).all_pix2world(w/2, h/2, 0, ra_dec_order=True)

def setCenterRADec(header, ra, dec):
    """
    Set the WCS reference point celestial coordinates.
    The reference point is in the image center.
    
    :param header: FITS header
    :param ra: in degrees within [0,360]
    :param dec: in degrees within [-90,90]
    """
    assert 0 <= ra <= 360
    assert -90 <= dec <= 90
    w = header['IMAGEW']
    h = header['IMAGEH']
    header['CRPIX1'] = int(w//2 + 1) # FITS is 1-based
    header['CRPIX2'] = int(h//2 + 1)
    header['CRVAL1'] = ra
    header['CRVAL2'] = dec
    
def readQuadMatch(fitsMatchPath):
    """
    
    :param fitsMatchPath: path to .match file written by astrometry.net
    :rtype: array of pixel coordinates of the stars in the quad, shape (n,2) in [x,y] order
    """
    with fits.open(fitsMatchPath) as hdulist:
        data = hdulist[1].data[0]
        starCount = data.field('DIMQUADS')
        starPxCoords = data.field('QUADPIX').reshape(-1,2)[:starCount]
        return starPxCoords
    
def readCorr(corrPath):
    """
    Return corresponding sources and reference stars as found by astrometry.net
    from the given .corr file.
    
    :param str corrPath: path to .corr file written by astrometry.net
    :rtype: tuple (xField, yField, xIndex, yIndex)
    """
    with fits.open(corrPath) as hdulist:
        data = hdulist[1].data
        xField = data.field('field_x')
        yField = data.field('field_y')
        xIndex = data.field('index_x')
        yIndex = data.field('index_y')
    
    return xField, yField, xIndex, yIndex

def readXy(fitsXyPath, sort=False, sortKey='FLUX', sortReverse=True, retSortField=False):
    """
    X,Y position of sources/stars with origin (0,0)
    
    .axy = extracted sources from image, includes flux
    .xyls = stars from reference catalog, no flux
    
    :raise KeyError: if sort is True and the sortKey doesn't exist 
    """
    with fits.open(fitsXyPath) as hdulist:
        data = hdulist[1].data
        x = data.field('X')-1 # FITS has (1,1) origin
        y = data.field('Y')-1
    if sort:
        flux = data.field(sortKey)
        fluxSort = np.argsort(flux)
        if sortReverse:
            fluxSort = fluxSort[::-1]            
        x = x[fluxSort]
        y = y[fluxSort]
    
    if sort and retSortField:
        return x,y,flux
    else:
        return x,y

def recomputeXylsPixelPositions(originalXylsPath, originalWcsPath, newWcsPathOrHeader):
    """
    Return pixel coordinates valid for `newWcsPathOrHeader` for
    the reference stars found in `originalXylsPath` (belonging to `originalWcsPath`).
    
    :rtype: tuple (x,y) with x and y being ndarrays
    """
    # Step 1: compute RA,Dec of reference stars (as this is not stored in .xyls)
    originalWCS = WCS(readHeader(originalWcsPath))
    x, y = readXy(originalXylsPath)
    ra, dec = originalWCS.all_pix2world(x, y, 0)
    
    # Step 2: compute pixel positions of reference stars in new WCS solution
    if isinstance(newWcsPathOrHeader, string_types):
        newWCS = WCS(readHeader(newWcsPathOrHeader))
    else:
        newWCS = WCS(newWcsPathOrHeader)

    # all_world2pix raised a NoConvergenc error
    # As we don't use SIP, we don't need to use all_world2pix.
    # wcs_world2pix doesn't support any distortion correction. 
    xNew, yNew = newWCS.wcs_world2pix(ra, dec, 0)
    
    return xNew,yNew

def getCatalogStars(header, limit=500, maxVmag=None, retVmag=False, retry=1):
    """
    Queries the Vizier catalog and retrieves stars for the sky area
    as defined by the given WCS header.
    
    :param header: FITS WCS header, must include IMAGEW and IMAGEH
    :param limit: maximum number of stars to return (optional)
    :param maxVmag: maximum magnitude of stars (optional)
    :param retVmag: if true, include Vmag in the result tuple
    :param retry: how many times to retry in case of errors (e.g. network problems)
    :rtype: tuple (x, y) or (x, y, vmag) sorted by decreasing brightness, origin (0,0)
            Note that vmag is a masked array and can contain masked values.
    """
    column_filters = {}
    if maxVmag:
        column_filters['VTmag'] = '<' + str(maxVmag)
        
    w, h = header['IMAGEW'], header['IMAGEH']
                
    # Step 1: query stars in tycho-2 online catalog, ordered by Vmag
    # We add a small border here. This is useful for
    # circling stars in an image, such that half circles
    # are drawn at the image corners instead of suddenly disappearing
    # circles.
    catalog = 'I/259/tyc2'
    centerRa, centerDec = getCenterRADec(header)
    border = 0.01 * w
    radiusBorder = getPixelScale(header)*border
    
    radius = getRadius(header) + radiusBorder
    if limit:
        # we have to query more stars as our search region is a circle
        # and we are filtering stars out afterwards
        row_limit = limit + int(limit*1.4)
    else:
        row_limit = -1
    print('Querying Vizier...')
    v = Vizier(columns=['_RAJ2000', '_DEJ2000', '+VTmag'],
               column_filters=column_filters, 
               row_limit=row_limit)
    try:
        result = v.query_region(coord.SkyCoord(ra=centerRa, dec=centerDec,
                                               unit=(u.deg, u.deg),
                                               frame='icrs'), 
                                radius=radius*u.deg, catalog=catalog)[0]
    except Exception as e:
        if retry > 0:
            print(repr(e))
            print('retrying...')
            time.sleep(2)
            # astroquery may have stored a corrupt response in its cache,
            # so we try again without using the cache
            # see https://github.com/astropy/astroquery/issues/465
            with suspend_cache(Vizier):
                return getCatalogStars(header, limit, maxVmag, retVmag, retry-1)
        print('Vizier query_region: ra={}, dec={}, radius={}, column_filters={}, row_limit={}, catalog={}'.
              format(centerRa, centerDec, radius, column_filters, row_limit, catalog),
              file=sys.stderr)
        raise
        
    vmag = result['VTmag']
    ra = result['_RAJ2000']
    dec = result['_DEJ2000']
    
    print(len(vmag), 'stars received')
    
    # Step 2: compute pixel coordinates for stars
    wcs = WCS(header)
    x, y = wcs.wcs_world2pix(ra, dec, 0)
    
    # Step 3: remove stars outside the image bounds
    # As above, we leave a small border here.
    inside = (-border <= y) & (y < h+border) & (-border <= x) & (x < w+border)
    x = x[inside]
    y = y[inside]
    vmag = vmag[inside]
    print(len(vmag), 'stars left after filtering')
    
    # Step 4: apply limit by removing the faintest stars
    if limit:
        x = x[:limit]
        y = y[:limit]
        vmag = vmag[:limit]
        
    if retVmag:
        return x, y, vmag
    else:
        return x, y
    
def writeXyls(path, x, y, vmag=None, clobber=False):
    """
    Writes an .xyls file as produced by astrometry.net.
    The input data can be retrieved using getCatalogStars.
    
    :param path: path to 
    :param x: x coordinates of stars, origin 0
    :param y: y coordinates of stars, origin 0
    :param vmag: if masked array, then masked values become nan
    :param clobber: overwrite output file if it already exists
    """
    assert len(x) == len(y)
    if vmag is not None:
        assert len(x) == len(vmag)
    x = x+1
    y = y+1
    colX = fits.Column(name='X', format='1D', array=x)
    colY = fits.Column(name='Y', format='1D', array=y)
    cols = [colX, colY]
    if vmag is not None:
        colVmag = fits.Column(name='Vmag', format='1D', array=vmag)
        cols.append(colVmag)
    tbhdu = fits.BinTableHDU.from_columns(cols)
    prihdr = fits.Header()
    prihdr['AN_FILE'] = ('XYLS', 'Astrometry.net file type')
    prihdu = fits.PrimaryHDU(header=prihdr)
    thdulist = fits.HDUList([prihdu, tbhdu])
    thdulist.writeto(path, checksum=True, clobber=clobber)
    
def getNoradId(header):
    """
    Returns the NORAD ID as a five-char string from the NORADID card of a FITS header,
    or None if the card doesn't exist.
    """
    noradId = header.get('NORADID')
    if noradId is not None:
        noradId = int(noradId)
    return noradId
    
def setNoradId(header, noradId):
    noradId = str(int(noradId))
    
    if header.get('NORADID') is None:
        header['HISTORY'] = 'NORADID added by auromat Python library'
    
    header['NORADID'] = (noradId, 'NORAD ID of spacecraft')

def getPhotoTime(header):
    """
    Return the phototime as found in DATE-OBS or None if not existing.
    
    :rtype: datetime.datetime | None
    """
    dateobs = header.get('DATE-OBS')
    if dateobs is None:
        return None
    else:
        try:
            d = datetime.strptime(dateobs, '%Y-%m-%dT%H:%M:%S.%f')
        except:
            d = datetime.strptime(dateobs, '%Y-%m-%dT%H:%M:%S')
        return d

def getShiftedPhotoTime(header):
    """
    Returns the corrected photo time or if not available, the
    original photo time.
    
    :rtype: datetime.datetime | None
    """
    _, d, _ = getShiftedSpacecraftPosition(header)
    if d is None:
        d = getPhotoTime(header)
    return d

def getSpacecraftPosition(header):
    """
    Returns the spacecraft position at the original photo time
    in cartesian GCRS coordinates.
    :rtype: tuple([x,y,z], date)
    """
    date = getPhotoTime(header)
    x = header.get('POSX')
    if x is None or date is None:
        return None, None
    y = header['POSY']
    z = header['POSZ']
    return np.array([x,y,z]), date

def setSpacecraftPosition(header, xyz, date):
    """
    Add the spacecraft position at the original photo time
    in cartesian GCRS coordinates to header cards
    POSX, POSY, POSZ, DATE-OBS.
    
    :param ndarray xyz: [x,y,z] position of spacecraft at 'date'
    :param datetime date: the original photo time
    """
    x,y,z = xyz
       
    if header.get('POSX') is None:
        header['HISTORY'] = 'POS* & DATE-OBS added by auromat Python library'
        
    header['POSX'] = (x, 'X coordinate of spacecraft in GCRS at DATE-OBS')
    header['POSY'] = (y, 'Y coordinate of spacecraft in GCRS at DATE-OBS')
    header['POSZ'] = (z, 'Z coordinate of spacecraft in GCRS at DATE-OBS')
    dateStr = date.strftime('%Y-%m-%dT%H:%M:%S.%f')
    header['DATE-OBS'] = (dateStr, 'EXIF timestamp of the photograph')

def getShiftedSpacecraftPosition(header):
    """
    Returns the spacecraft position at the corrected photo time
    in cartesian GCRS coordinates.
    :rtype: tuple([x,y,z], datetime date, timedelta delta)
    """
    date = getPhotoTime(header)
    shift = header.get('DATESHIF')
    x = header.get('POSXSHIF')
    if x is None or date is None or shift is None:
        return None, None, None
    y = header['POSYSHIF']
    z = header['POSZSHIF']
    delta = timedelta(seconds=shift)
    shiftedDate = date + delta
    return np.array([x,y,z]), shiftedDate, delta
    
def setShiftedSpacecraftPosition(header, xyz, shiftedDate):
    """
    Add the spacecraft position at the given shifted photo time
    in cartesian GCRS coordinates to header cards
    POSXSHIF, POSYSHIF, POSZSHIF, DATESHIF.
    
    :param ndarray xyz: [x,y,z] position of spacecraft at 'date'
    :param datetime shiftedDate: the shifted photo time
    """
    x,y,z = xyz
    
    date = getPhotoTime(header)
    if date is None:
        raise ValueError('DATE-OBS needs to be set before setting the shifted date')
    delta = (shiftedDate - date).total_seconds()
    
    if header.get('POSXSHIF') is None:
        header['HISTORY'] = 'POS*SHIF & DATESHIF added by auromat Python library'
        
    header['POSXSHIF'] = (x, 'X coordinate of spacecraft in GCRS at DATESHIF')
    header['POSYSHIF'] = (y, 'Y coordinate of spacecraft in GCRS at DATESHIF')
    header['POSZSHIF'] = (z, 'Z coordinate of spacecraft in GCRS at DATESHIF')
    header['DATESHIF'] = (delta, 'DATE-OBS shift in seconds')



# Copyright European Space Agency, 2013

from __future__ import division

import unittest
from nose.plugins.attrib import attr
import os
import tempfile

from auromat.mapping.spacecraft import getMapping
import auromat.export.cdf
from spacepy import pycdf
from auromat.mapping.cdf import CDFMapping
from auromat.test.export_netcdf_test import check_equal

@attr('slow')
class Test(unittest.TestCase):

    def testCDFExport(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
        path = tempfile.mktemp(suffix='.cdf')
        try:            
            auromat.export.cdf.write(path, mapping)
            
            with pycdf.CDF(path) as root:
                assert set(root.keys()) == set(['Epoch', 'lat', 'lon', 'altitude',
                                                'lat_bounds', 'lon_bounds', 'mlat',
                                                'mlt', 'mlat_bounds', 'mlt_bounds',
                                                'mcrs', 'img_red', 'img_green',
                                                'img_blue', 'zenith_angle', 'camera_pos',
                                                'crs'])
            
            cdfmapping = CDFMapping(path)
            cdfmapping.checkGuarantees()
            check_equal(cdfmapping, mapping)
            
        finally:
            if os.path.exists(path):
                os.remove(path)
                        
            
def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)
